%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CRAY-CMAC-2.SL - CRAY PSL cmacros Part 2                              %
%                                                                       %
% Author:      M. L. Griss, S. Lowder, R. Kessler                       %
% (PSL3.2)     Computer Science Dept.                                   %
%              University of Utah                                       %
% Date:        1 June 1982                                              %
%                                                                       %
% Based on M68K-CMAC.SL  1 June by Lowder et al                         %
%                                                                       %
% Revisions    changed nearly all for Cray PSL 3.4                      %
%              by H. Melenk and W. Neun,ZIB Berlin                      %
%                                                                       %
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%========================================================================
%   Initially, we define a number of macros that will be used by other
%    routines.
%---------------------------------------------------------------------
%   *3Op handles the tertiary opcodes of the CRAY.  It is used for all
%    of the opcodes that look the same, i.e. AND, OR, XOR, ADD and SUB.
%
%   First comes the DE to make the Cmacro real.
%   --- We have to go through all of this crap to get the CMACRO to
%       allow us to pass in the CAL opcode in the CMACRO.  This should
%       be modified so it is easier to do this^^^
 
(De *3op (ArgOne ArgTwo ArgThree Instruction)
 (prog (ResultingCode*)
    (return (CMacroPatternExpand
              (list (ResolveOperand '(Reg t1)     ArgOne)
                    (ResolveOperand '(Reg t2)     ArgTwo)
                    (ResolveOperand '(Reg t3)     ArgThree)
                    Instruction)
              (get '*3op 'CMacroPatternTable)))))
 
%   FORMAT: Destination Operand1 Operand2 Operator
 
(DefCMacro *3Op
      ((AnyP WZeroP AnyP AddOperatorP)   (*Move ArgThree ArgOne))
      ((AnyP AnyP WZeroP ARegOperatorP)  (*Move ArgTwo ArgOne))
 
      ((ARegP WZeroP ARegP SubOperatorP) (ArgFour ArgOne ArgTwo ArgThree))
      ((SRegP WZeroP SRegP SubOperatorP) (ArgFour ArgOne (Reg S0) ArgThree))
 
      ((ARegP ARegP ARegP ARegOperatorP) (ArgFour ArgOne ArgTwo ArgThree))
      ((ARegP ARegP OneP ARegOperatorP)  (ArgFour ArgOne ArgTwo ArgThree))
      ((ARegP OneP ARegP AddOperatorP)   (ArgFour ArgOne ArgThree ArgTwo))
      ((ARegP BRegP OneP ARegOperatorP)  (*Move ArgTwo (Reg A6))
                                         (ArgFour ArgOne (Reg A6) ArgThree))
      ((ARegP OneP BRegP AddOperatorP)   (*Move ArgThree (Reg A6))
                                         (ArgFour ArgOne (Reg A6) ArgTwo))
      ((ARegP AnyARegButA7P AnyP ARegOperatorP) (*Move ArgThree (Reg A7))
                        (*3Op ArgOne ArgTwo (Reg A7) ArgFour))
      ((ARegP ARegP AnyP ARegOperatorP)  (*Move ArgThree (Reg A6))
                        (*3Op ArgOne ArgTwo (Reg A6) ArgFour))
      ((ARegP AnyP AnyARegButA7P ARegOperatorP)  (*Move ArgTwo (Reg A7))
                        (*3Op ArgOne (Reg A7) ArgThree ArgFour))
      ((ARegP AnyP ARegP ARegOperatorP)  (*Move ArgTwo (Reg A6))
                        (*3Op ArgOne (Reg A6) ArgThree ArgFour))
      ((ARegP AnyP AnyP ARegOperatorP)    (*Move ArgTwo (Reg A6))
                                          (*Move ArgThree (Reg A7))
                        (*3Op ArgOne (Reg A6) (Reg A7) ArgFour))
      ((BRegP AnyP AnyP ARegOperatorP)
                        (*3Op (Reg A6) ArgTwo ArgThree ArgFour)
                        (*Move (Reg A6) ArgOne))
      ((SRegP SRegP SRegP) (ArgFour ArgOne ArgTwo ArgThree))
      ((SRegP AnyP  AnySRegButS7P) (*Move ArgTwo (Reg S7))
                        (*3Op ArgOne (Reg S7) ArgThree ArgFour))
      ((SRegP AnyP  SRegP) (*Move ArgTwo (Reg S6))
                        (*3Op ArgOne (Reg S6) ArgThree ArgFour))
      ((SRegP AnySRegButS7P AnyP) (*Move ArgThree (Reg S7))
                        (*3Op ArgOne ArgTwo (Reg S7) ArgFour))
      ((SRegP SRegP AnyP)  (*Move ArgThree (Reg S6))
                        (*3Op ArgOne ArgTwo (Reg S6) ArgFour))
      ((SRegP AnyP  AnyP)  (*Move ArgTwo (Reg S6))
                        (*Move ArgThree (Reg S7))
                        (*3Op ArgOne (Reg S6) (Reg S7) ArgFour))
      (                 (*3Op (Reg S6) ArgTwo ArgThree ArgFour)
                        (*Move (Reg S6) ArgOne)))
 
%---------------------------------------------------------------------
%   The sign extension cmacro.  Signextend the argument, either by a literal
%   number or the contents of an A Register.  The bit position is Labeled
%   from the left (position 0).
%
%   First we define the procedure to work as a 2 operand cmacro.
(de *SignExtend (ArgOne ArgTwo)
   (Expand2OperandCmacro ArgOne ArgTwo '*SignExtend))
 
%   FORMAT: LocationToSignExtend BitPositionToExtend
%:::::::::: WARNING - DOESN'T work on non integer bit position.
(DefCMacro *SignExtend
     ((AnySRegButS7P SmallInumP) (*Move (Difference 63 ArgTwo) (Reg A6))
                     (DRshift (Reg S7) ArgOne (Reg A6))
                     (Rshift ArgOne ArgOne (Difference 63 ArgTwo))
                     (MoveNEG ArgOne ArgOne)
                     (DLshift ArgOne (Reg S7) (Reg A6)))
     ((SRegP SmallInumP)      (*Move (Difference 63 ArgTwo) (Reg A6))
                     (DRshift (Reg S6) ArgOne (Reg A6))
                     (Rshift ArgOne ArgOne (Difference 63 ArgTwo))
                     (MoveNEG ArgOne ArgOne)
                     (DLshift ArgOne (Reg S6) (Reg A6)))
    (                (*Move ArgOne (Reg S6))
                     (*SignExtend (Reg S6) ArgTwo)
                     (*Move (Reg S6) ArgOne)))
 
%---------------------------------------------------------------------
%   A new Cmacro to strip off the tag and place it non sign extended
%    into the argument
 
%   First we define the procedure to work as a 2 operand cmacro.
 
(de *Tag (ArgOne ArgTwo)  (Expand2OperandCmacro ArgOne ArgTwo '*Tag))
 
%   FORMAT: LocationToPlaceResult ArgumentToStripTag
 
(DefCMacro *Tag
         ((SRegP AnyP)(*Move ArgTwo ArgOne)
                      (*Wshift ArgOne (Minus NonTagPart)))
         ((AnyP AnySRegButS7P)(*Move ArgTwo (Reg S7))
                      (*Wshift (Reg S7) (Minus NonTagPart))
                      (*Move (Reg S7) ArgOne))
         (            (*Move ArgTwo (Reg S6))
                      (*Wshift (Reg S6) (Minus NonTagPart))
                      (*Move (Reg S6) ArgOne)))
 
%---------------------------------------------------------------------
%   Now the tertiary operand macros.
%   FORMAT: Destination Operand1 Operand2
 
(de *WAND3 (Destination Operand1 Operand2)
            (*3Op Destination Operand1 Operand2 'AND))
(DefCMacro *WAND3)
 
(de *WDIFFERENCE3 (Destination Operand1 Operand2)
            (*3Op Destination Operand1 Operand2 'SUB))
(DefCMacro *WDIFFERENCE3)
 
(de *WOR3 (Destination Operand1 Operand2)
            (*3Op Destination Operand1 Operand2 'OR))
(DefCMacro *WOR3)
 
(de *WPLUS3 (Destination Operand1 Operand2)
            (*3Op Destination Operand1 Operand2 'ADD))
(DefCMacro *WPLUS3)
 
(de *WXOR3 (Destination Operand1 Operand2)
         (cond
            ((Wzerop operand1) (*Move Operand2 Destination))
            ((Wzerop operand2) (*Move Operand1 Destination))
            (t (*3Op Destination Operand1 Operand2 'XOR))
)        )
 
(DefCMacro *WXOR3)
 
%========================================================================
%   The following are each of the real CMACRO's as needed by the compiler.
 
%---------------------------------------------------------------------
%   *Alloc:Allocate N locations on the stack.
%   CRAY Decisions - In order to use the Crays Return Jump instruction,
%    we must save and restore the B00 Register.  Therefore, we will
%    save it upon every alloc and restore in at dealloc time.
%   FORMAT: Number-of-Locations
 
% The strategy is to place the top stack frame into the T Registers.
% This will allow fast frame access, and also moving the stack into
% memory can be accomplished with a fast block Move instruction.  The T
% Regs will be allocated as follows:
 
%       T70 - Top frame (frame 33)
%       T69 - Top -1 (frame 32)
%       ....
%       T31 - Second Frame Element (frame 2)
%       T30 - First Frame Element (frame 1)
%       T27 - Return Address (caller's B00)
%       T26 - Previous Stack Frame Size (padded)
%
% Alternatively : if no links are present and a small allocation
%                 of frame elements is needed (<6) these elements
%                 are allocated in T71-T75.
%                 So no Frame must be saved (Fast-Frame feature)
%  Feb-86 ,ZIB Berlin
% Alternatively : if links are present and a no allocation
%                 of frame elements is needed (*alloc 0)
%                 the function will save its B00 and A3 itself
%                 So no Frame must be saved by called functions
%  Apr-86 ,ZIB Berlin
 
% This assumes that the A3 Register coming into this function has the
% size of the current frame in the TRegs.
%
% Checking the stackpointer against Stackupperbound, which is a BReg ^
 
(DefCMacro *Alloc              % Stack grows UP
         ((Fast-Frame))
         ((Freep))
         ((Zerop)
             (*Move (Reg st)  (Reg a0))    % Sets up address for save
             (blksto (Reg a3) (Reg PrevFrameSize)) % Save to memory
             (*wplus2 (Reg st) (Reg a3))   % Update stack pointer
             (*Move (Reg stkupbnd) (Reg a6)) % CHECK STACK ^
             (*wdifference3 (Reg a0) (Reg st) (Reg a6))
             (*Move (Reg b00) (Reg a6))    % Fetch Return addr to TReg
             (store (Reg a3) 0 (Reg st))   % First to Stack + 0
             (store (Reg a6) 1 (Reg st))   % First to Stack + 1
             (*wplus2 (Reg st) 2)          % Update stack pointer
             (*Move 0 (Reg a3))            % Set up current size
             (jap (plus2 (parcel SYMFNC)
                         (times2 (idloc Stack-Blown-Up) 4))))
 
         ((FitsInStackP)
             (*Move (Reg st)  (Reg a0))    % Sets up address for save
             (blksto (Reg a3) (Reg PrevFrameSize)) % Save to memory
             (*wplus2 (Reg st) (Reg a3))   % Update stack pointer
             (*Move (Reg stkupbnd) (Reg a1)) % CHECK STACK !
             (*wdifference3 (Reg a0) (Reg st) (Reg a1))
             (*Move (Reg a3) (Reg s7))     % Save old frame size
             (*Move (Reg b00) (Reg a6))    % Fetch Return addr to TReg
             (*Move (Reg a6) (Reg s6))     % First to SReg
             (*Move (plus2 ArgOne 2) (Reg a3))  % Set up current size
             (jap (plus2 (parcel SYMFNC)
                         (times2 (idloc Stack-Blown-Up) 4)))
             (*Move (Reg s6) (Reg ReturnAddr))  % Finally Store B00
             (*Move (Reg s7) (Reg PrevFrameSize)))
 
         (   (*Cerror "Alloc of too large a frame" 'ALLOC ArgOne)))
 
%---------------------------------------------------------------------
%
(de *useFastFrame (x) nil)
(defcmacro *useFastFrame)
 
%---------------------------------------------------------------------
%   Arg1 <- Arithmeticshift Arg1 by Arg2 bit positions.  + Arg2 to left,
%    - Arg2 to the right.  It does sign extension.
%   FORMAT: Source-Destination BitPositions
 
(DE *Ashift (ArgOne ArgTwo) (EXPAND2OPERANDCMACRO ArgOne ArgTwo '*Ashift))
 
(DefCMacro *Ashift
         ((AnyP  WZeroP))
         ((SRegP PosLE64P)   (*Wshift ArgOne ArgTwo))
         ((SRegP NegLE64P)   (*Wshift ArgOne ArgTwo)
                             (*SignExtend ArgOne (Minus ArgTwo)))
         ((SRegP ARegP)      (*JumpWlessP TempLabel ArgTwo 0)
                             (Lshift ArgOne ArgOne ArgTwo)
                             (*JUMP TempLabel2)
                             (*Lbl (Label TempLabel))
                             (*WMINUS ArgTwo ArgTwo)
                             (Rshift ArgOne ArgOne ArgTwo)
                             (*SignExtend ArgOne ArgTwo)
                             (*Lbl (Label TempLabel2)))
         ((SRegP)            (*Move ArgTwo (Reg A6))
                             (*Ashift ArgOne (Reg A6)))
         (                   (*Move ArgOne (Reg S6))
                             (*Ashift (Reg S6) ArgTwo)
                             (*Move (Reg S6) ArgOne)))
 
%---------------------------------------------------------------------
%   Call the argument.
%   FORMAT: AddressToCall
(DefCMacro *Call
     ((InternallyCallableP) (R (InternalEntry ArgOne)))
     (          (R (plus2 (parcel SymFnc) (times2 4 (idLoc ArgOne))))))
 
 
(De *Cerror (x a b)
  (Prog (I)
        (Setq I (Wrs NIL))
        (Printf "%n *** Cerror: %r %r %r %n " X A B)
        (Wrs I)
        (Return (List (List 'Cerror x a b)))))
 
(DefCMacro *Cerror)
%---------------------------------------------------------------------
%   Deallocate argument elements off the stack
%   FORMAT: Number-of-locations
%  Warning: Number-of-locations must be a constant.
(DefCMacro *DeAlloc
         ((Fast-Frame-Aux))
         ((Freep))
         ((Zerop)
             (load (Reg a3) -2 (Reg st))
             (load (Reg a6) -1 (Reg st))
             (*Move 2 (Reg a7))
             (*wdifference (Reg st) (Reg a7))
             (*wdifference (Reg st) (Reg a3))    % Take the Stack down
             (*Move (Reg st) (Reg a0))           % Set up address
             (blklod (Reg a3) (Reg PrevFrameSize))% Do the load
             (*Move (Reg a6) (Reg b00))
             )
         (   (*Move (Reg PrevFrameSize) (Reg s7))% And Prev Frame ->a3
             (*Move (Reg s7) (Reg a3))           % Finally into a3
             (*wdifference (Reg st) (Reg a3))    % Take the Stack down
             (*Move (Reg ReturnAddr) (Reg s6))   % Move Retaddr to B00
             (*Move (Reg st) (Reg a0))           % Set up address
             (blklod (Reg a3) (Reg PrevFrameSize))% Do the load
             (*Move (Reg s6) (Reg a6))           % Finish Retaddr Move
             (*Move (Reg a6) (Reg B00))
 )        )        % We are done
 
(De *PreDeAlloc(x)
   (cond ((Fast-Frame-Aux x) nil)
         ((Freep x) nil)
         ((Zerop x) nil)
         (t '((*move (reg PrevFrameSize) (Reg S6))
              (*move (reg ReturnAddr) (Reg S7))
              (Store (Reg S6) 0 (Reg st))
              (Store (Reg S7) 1 (Reg st))
              (*Move 2 (Reg a7))
              (*wplus2 (Reg st) (Reg a7))
              (*Move 0 (Reg a3))))
 ) )
(defcmacro *PreDeAlloc)
 
(De *PostDeAlloc(x)
   (cond ((Fast-Frame-Aux x) nil)
         ((Freep x) nil)
         (t '((load (Reg a3) -2 (Reg st))
              (load (Reg a6) -1 (Reg st))
     %
     % on X-MP a (BLKlod for b00 and b01 could be better (b01-> a3)
     %
              (*Move 2 (Reg a7))
              (*wdifference (Reg st) (Reg a7))
              (*wdifference (Reg st) (Reg a3))    % Take the Stack down
              (*Move (Reg st) (Reg a0))           % Set up address
              (blklod (Reg a3) (Reg PrevFrameSize))% Do the load
              (*Move (Reg a6) (Reg b00))))
 ) )
(defcmacro *PostDeAlloc)
 
%---------------------------------------------------------------------
%   Return jump
(De *End(x)
             '(    (J (Reg B00))))
 
(defcmacro *End)
 
%---------------------------------------------------------------------
%   Deallocate argument elements off the stack and return
%   FORMAT: Number-of-Locations
(DefCMacro *Exit
         (          (*DeAlloc ArgOne)
                    (J (Reg B00))))
 
%---------------------------------------------------------------------
%   Extract the Field from the Operand, From the BitPosition of Length
%    into the Destination.
%   FORMAT: Destination Operand BitPosition Length
%?????? This assumes that the bitpositions and length are constants.  Actually
%??????  will still work for bitposition a variable (I think).
(DefCMacro *Field
         ((AnyP AnyP WZeroP WordWidthP)  (*Move ArgTwo ArgOne))
         ((SRegP)       (*Move ArgTwo ArgOne)
                        (*Wshift ArgOne ArgThree)
                        (*Wshift ArgOne (difference ArgFour 64)))
         ((AnyP AnySRegButS7P)(*Field (Reg S7) ArgTwo ArgThree ArgFour)
                        (*Move (Reg S7) ArgOne))
         % Next pattern knows that second arg is (Reg S7)
         ((AnyP SRegP)  (*Field (Reg S6) ArgTwo ArgThree ArgFour)
                        (*Move (Reg S6) ArgOne))
         (              (*Field (Reg S7) ArgTwo ArgThree ArgFour)
                        (*Move (Reg S7) ArgOne)))
 
%---------------------------------------------------------------------
%   Used for LINKE (non-return) calls
%   FORMAT: Address-to-call
(DefCMacro *JCall
    ((InternallyCallableP)  (J (InternalEntry ArgOne)))
    (        (J (plus2 (parcel SymFnc) (times2 4 (idLoc ArgOne))))))
 
% support for optimization of never come back functions
% in Pass 2 (&call)
 
(De *JCall-never-return    (functionname NumberofArgs)
      (append (MakeLinkRegs FunctionName NumberofArgs)
              (list (list '*JCall FunctionName))))
 
(DefCMacro *Jcall-never-return)
 
%---------------------------------------------------------------------
%   Jump to the argument's address
%   FORMAT: Address
(DefCMacro *Jump
         ((TaggedLabel)      (J ArgOne))
%  We assume for the default case that the argument is actually a word
%   address, so we must shift it left by 2 to turn into a parcel address.
%   Note, that B Registers aren't good for a whole lot, so we must work with
%   an S Register.
%  It looks like this will always be an indexed operation from memory.
%   So, what we need to do is compute the index our selves and then
%   shift left by 2 and then branch.  What this is supposed to do, is
%   The indexed will turn into some instructions (because of its
%   anyReg property) and then will be (Indexed (Reg xx) 0).  We
%   have the address in (Reg xx), so Move into S6, shift and branch.
         (          (*Move (PickOffIndexedReg ArgOne) (Reg S6))
                    (*Wshift (Reg S6) 2)
                    (*Move (Reg S6) (Reg B02))
                    (J (Reg B02))))
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the operands are eq.
%   FORMAT: Address Operand1 Operand2
(DefCmacro *JumpEq
         ((ABRegP ABRegP)     (*WDifference3 (Reg A0) ArgOne ArgTwo)
                              (JAZ ArgThree))
         ((ABRegP SmallInumP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                              (JAZ ArgThree))
         ((SmallInumP ABRegP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                              (JAZ ArgThree))
         (                    (*WXor3 (Reg S0) ArgOne ArgTwo)
                              (JSZ ArgThree)))
 
%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address (SRegJumpOpCode. ARegJumpOpCode)
(DefCMacro *JumpIf
         ((ABRegP ABRegP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                          (ArgFive ArgThree))
         ((ABRegP SmallInumP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                                    (ArgFive ArgThree))
         ((SmallInumP ABRegP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                                    (ArgFive ArgThree))
         (              (*WDifference3 (Reg S0) ArgOne ArgTwo)
                        (ArgFour ArgThree)))
 
(de *JumpIF (ArgOne ArgTwo Label Instructions)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
               (list (ResolveOperand '(Reg t1) ArgOne)
                     (ResolveOperand '(Reg t2) ArgTwo)
                     (ResolveOperand '(Reg Error) Label)
                     (car Instructions)
                     (cdr Instructions))
                     (get '*JumpIF 'CMacroPatternTable)))))
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is in the range,
%    from 0 to TypeTag or is a Negative Integer Tag (-1).
%   Be careful, since the operand should not be destroyed^^^
%:::::::::: Weirdee CMACRO type checking is different from call^^^^????
%   FORMAT: Operand TypeTag Address
(DefCmacro *jumpintype
    ((TempSregp) (*TAG (Reg S7) ArgOne)
                 (*JUMPWLEQ ArgThree (Reg S7) ArgTwo)
                 (*JUMPEQ ArgThree (Reg S7) 31))
    ((anyp wzerop)                            % intp case
                 (*Move ArgOne (Reg S6))
                 (*TAG (Reg S6)(Reg S6))
                 (*JUMPEQ ArgThree (Reg S6) 0)
                 (*JUMPEQ ArgThree (Reg S6) 31))
    (            (*Move ArgTwo (Reg S7))
                 (*Move ArgOne (Reg S6))
                 (*TAG (Reg S6)(Reg S6))
                 (*JUMPWLEQ ArgThree (Reg S6) (Reg S7))
                 (*JUMPEQ ArgThree (Reg S6) 31))
)
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the operands are not eq.
%   FORMAT: Address Operand1 Operand2
(DefCmacro *JumpNotEQ
         ((ABRegP ABRegP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                                (JAN ArgThree))
         ((ABRegP SmallInumP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                                (JAN ArgThree))
         ((SmallInumP ABRegP) (*WDifference3 (Reg A0) ArgOne ArgTwo)
                                (JAN ArgThree))
         (              (*WXor3 (Reg S0) ArgOne ArgTwo)
                                (JSN ArgThree)))
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is not in the range,
%    from 0 to TypeTag or is not a Negative Integer Tag (-1).
%   Be careful, since the operand should not be destroyed !
%:::::::::: Weirdee CMACRO type checking is different from call^^^^????
%   FORMAT: Operand TypeTag Address
(DefCmacro *jumpNOTintype
    ((TempSregp) (*TAG (Reg S7) ArgOne)
                 (*JUMPWLEQ TempLabel (Reg S7) ArgTwo)
                 (*JUMPnotEQ ArgThree (Reg S7) 31)
                 (*Lbl (Label TempLabel)))
    ((anyp Wzerop)                       %intp test
                 (*Move ArgOne (Reg S7))
                 (*TAG (Reg S7)(Reg S7))
                 (*JumpEq TempLabel (Reg S7) 0)
                 (*JumpnotEq ArgThree (Reg S7) 31)
                 (*Lbl (Label TempLabel)))
    (            (*Move ArgTwo (Reg S6))
                 (*Move ArgOne (Reg S7))
                 (*TAG (Reg S7)(Reg S7))
                 (*JumpWLEQ TempLabel (Reg S7) (Reg S6))
                 (*JumpnotEQ ArgThree (Reg S7) 31)
                 (*Lbl (Label TempLabel)))
)
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is not eq to
%    the TypeTag.
%   Be careful, since the operand should not be destroyed^^^
%:::::::::: Weirdee CMACRO type checking is different from call^^^^????
%   FORMAT: Operand TypeTag Address
(DefCMacro *JumpNotType
       ((TempSregp) (*Tag (Reg S7) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S7) ArgTwo))
       ((Anyp Wzerop) (*Tag (Reg S6) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S6) 0))
       (            (*move ArgTwo (reg S7))
                    (*Tag (Reg S6) ArgOne)
                    (*JumpNotEQ ArgThree (Reg S6) (reg s7))))
 
%---------------------------------------------------------------------
%   Jump to the argument's address if the Tag(Operand) is eq to
%    the TypeTag.
%   Be careful, since the operand should not be destroyed^^^
%:::::::::: Weirdee CMACRO type checking is different from call^^^^????
%   FORMAT: Operand TypeTag Address
(DefCMacro *JumpType
          ((TempSregp)  (*Tag (Reg S7) ArgOne)
                        (*JumpEQ ArgThree (Reg S7) ArgTwo))
          ((Anyp Wzerop)(*Tag (Reg S6) ArgOne)
                        (*JumpEQ ArgThree (Reg S6) 0))
          (             (*move Argtwo (reg S7))
                        (*Tag (Reg S6) ArgOne)
                        (*JumpEQ ArgThree (Reg S6) (reg S7))))
 
%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 >= Operand2
%   FORMAT: Address Operand1 Operand2
 
(De *JumpWGEQ (Lbl ArgOne ArgTwo)
                   (*JumpIf ArgOne ArgTwo Lbl '(JSP . JAP)))
(DefCmacro *JumpWGEQ)
 
%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 > Operand2
%   FORMAT: Address Operand1 Operand2
 
(De *JumpWGreaterP (Lbl ArgOne ArgTwo)
                   (*JumpIf ArgTwo ArgOne Lbl '(JSM . JAM)))
(DefCmacro *JumpWGreaterP)
 
%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 <= Operand2
%   FORMAT: Address Operand1 Operand2
 
(De *JumpWLEQ (Lbl ArgOne ArgTwo)
             (*JumpIf ArgTwo ArgOne Lbl '(JSP . JAP)))
(DefCmacro *JumpWLEQ)
 
%---------------------------------------------------------------------
%   Jump to the argument's address if Operand1 < Operand2
%   FORMAT: Address Operand1 Operand2
 
(De *JumpWLessP (Lbl ArgOne ArgTwo)
       (*JumpIf ArgOne ArgTwo Lbl '(JSM . JAM)))
(DefCmacro *JumpWLessp)
 
%---------------------------------------------------------------------
%   FORMAT: LabelName
 
(DefCMacro *Lbl
         (                  ArgOne))
%---------------------------------------------------------------------
%   PSL34 macro preparing the FastApplyCall
 
(DefCMacro *fast-apply-load
         (            (*Move ArgOne (Reg t1))))
%-------------------------------------------------------------
%   PSL34 macro for IDAPPLY
%   by convention the id has been loaded by *fast-apply-load
%   the first parameter defines the number of arguments to
%   the function called by IDAPPLY
(defCmacro *Id-apply)
 
(de *Id-Apply (nargs)
    `(
      (*Move (Reg symfnc)(Reg a6))       % calculate jumpaddr
      (*Move (Reg t1)(Reg s7))
      (*Move (Reg s7)(Reg a7))
      (*wplus2 (Reg a6)(Reg a7))
      (*move (reg 3) (reg t1))
      (load (reg 3) (indexed (reg a6) 0))
      (*wshift (Reg s7) (wconst 4))      % &linkshiftbits&
      (*Move (wconst ,nargs)(Reg s6))    % calculate nargsReg
      (*wor (Reg s6)(Reg s7))
      (*Move (Reg s6)(Reg NargReg))
      (*wshift (reg 3) (wconst -32))
      (*move (reg 3) (reg a6))
      (*move (reg t1) (reg 3))
      (*retjump (Reg a6))
))
 
%-------------------------------------------------------------
%   PSL34 macro for addressapply
%   by convention the id has been loaded by *fast-apply-load
%   the first parameter defines the number of arguments to
%   the function called by addressAPPLY
(defCmacro *address-apply)
(de *address-Apply (nargs)
    `(
      (*Move (Reg t1)(Reg s7))
      (*Move (Reg s7)(Reg a6))
      (*Move (wconst ,nargs) (Reg s7))
      (*Move (Reg s7)(Reg nargReg))
      (*retjump (Reg a6))
))
 
%-----------------------------------------------------------------
%  (*RETJUMP Reg)
%   return jump to an address given in a Register
 
(defCmacro *retjump)
 
(DE *retjump (Register)
       (prog (LL cde)
           (setq cde
             (list
                (list 'ploc '(Reg a7) (list 'Label (setq ll (gensym))))
                '(Move (Reg B00) (Reg A7))
                (list '*Move Register '(Reg B07))
                '(J (Reg B07))
                (LIST '*Lbl (List 'Label LL))
          )  )
              (RETURN CDE)
)     )
 
%---------------------------------------------------------------------
%   FORMAT: FunctionName FunctionType NumberofArgs
 
(de *Link (FunctionName FunctionType NumberofArgs)
  (cond ((FlagP FunctionName 'ForeignFunction)
        `((*ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)))
        ((FlagP FunctionName 'C-ForeignFunction)
        `((*C-ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)))
        ((or (get FunctionName 'opencode)
             (get FunctionName 'exitopencode)
             (InternallyCallableP FunctionName))
                         `((*Call ,FunctionName)))
        (t  (append (MakeLinkRegs FunctionName NumberofArgs)
                         `((*Call ,FunctionName))))))
 
(DefCMacro *Link)
 
%---------------------------------------------------------------------
%   FORMAT: DeAllocCount FunctionName FunctionType NumberofArgs
 
(de *LinkE (DeAllocCount FunctionName FunctionType NumberofArgs)
   (cond ((FlagP FunctionName 'ForeignFunction)
          `((*ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)
            (*Exit ,DeAllocCount)))
         ((FlagP FunctionName 'C-ForeignFunction)
          `((*C-ForeignLink ,FunctionName ,FunctionType ,NumberofArgs)
            (*Exit ,DeAllocCount)))
         ((or (get FunctionName 'exitopencode)
              (InternallyCallableP FunctionName))
                    `((*DeAlloc ,DeAllocCount)
                      (*JCall ,FunctionName)))
         ((get FunctionName 'opencode)
                    `((*Call ,FunctionName)
                      (*exit ,DeAllocCount)))
         (t  (append (MakeLinkRegs FunctionName NumberofArgs)
                     `((*DeAlloc ,DeAllocCount)
                       (*JCall ,FunctionName))))))
 
(DefCMacro *LinkE)
 
(de *PostLinkE (DeAllocCount FunctionName FunctionType NumberOfArgs)
   (cond ((FlagP FunctionName 'ForeignFunction)
          `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArgs)
            (J (Reg B00))))
         ((FlagP FunctionName 'C-ForeignFunction)
          `((*C-ForeignLink ,FunctionName ,FunctionType ,NumberOfArgs)
            (J (Reg B00))))
         ((InternallyCallableP FunctionName)
                   `((*JCall ,FunctionName)))
         ((get FunctionName 'exitopencode)
                   `((*JCall ,FunctionName)))
         ((get FunctionName 'opencode)
                   `((*Call ,FunctionName)
                     (*end ,dealloccount)))
         (t (append (MakeLinkRegs FunctionName Numberofargs)
                     `((*JCall ,FunctionName))))
)  )
 
(DefCMacro *PostLinkE)
 
%---------------------------------------------------------------------
%  Fetch the address of the second operand into the destination
%   FORMAT: Destination ImmediateAddress
%   Dont use *LOC for ParcelAddresses, use *PLOC !!!!
 
(DefCMacro *Loc
        ((RegP IndexedP) (*Move (PickOffIndexedReg ArgTwo) (Reg A6))
                         (*WPLus2 (Reg A6) (PickOffIndexAmt ArgTwo))
                         (*Move (Reg A6) ArgOne))
        ((ASRegP AnyP)   (Move ArgOne ArgTwo))
        ((TRegP AnyP)    (*Move ArgTwo (Reg S6))
                         (*Move (Reg S6) ArgOne))
        (                (*Move ArgTwo (Reg A6))
                         (*Move (Reg A6) ArgOne)))
 
(defcmacro *ploc
       (      (ploc argone argtwo)))
 
(de *ploc (argone argtwo) (expand2operandcmacro argone argtwo '*ploc))
 
