%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PC:PASS-1-CRAY.SL
% Description:  CRAY Machine-dependent compiler declarations and functs
% Author:       Herbert Melenk, Winfried Neun
% Created:      4-Sep-86
% Mode:         Lisp
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1987, Konrad Zuse Zentrum, Berlin
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Revisions:
% W. Neun   20-Aug 1989: correct bug in pa1r-and which lost the values of and
% H. Melenk 20-Jul 1987: simpler pass1-macros for intp,fixp, numberp
 
%(compiletime (off lapopt))
(compiletime (ds string-concat ( a b)
                  (compress (append (explode2 a) (explode2 b)))))
(compiletime (de knowntobeinteger (x) nil))
 
(compiletime (load if-system))
 
(compiletime (flag '(mappand binaryandp binaryandoperandp
                     pass-1-test-ands
                     pass-1-flatten-ands
                     islocalref
             knowntobeinteger knowntobeintegeraux) 'internalfunction))
 
(compiletime
 (defmacro def-pass-1-function (name args . body)
   (let ((fcn-name (intern (string-concat "PA1F-" (id2string name)))))
     `(progn
       (put (quote ,name) (quote pass-1-function) (quote ,fcn-name))
       (de ,fcn-name ,args ,@body)
       ))))
 
(compiletime
 (defmacro def-pass-1-reform (name args . body)
   (let ((fcn-name (intern (string-concat "PA1R-" (id2string name)))))
     `(progn
       (put (quote ,name) (quote pass-1-reform) (quote ,fcn-name))
       (defmacro ,fcn-name ,args ,@body)
       ))))
 
(compiletime
 (defmacro def-pass-1-integer-reform (name args . body)
   (let ((fcn-name (intern (string-concat "PA1R-" (id2string name)))))
     `(progn
       (put (quote ,name) (quote pass-1-reform) (quote ,fcn-name))
       (flag (quote (,name)) (quote integer-transparent))
       (defmacro ,fcn-name ,args ,@body)
       ))))
 
(def-pass-1-reform  aplus2 (argone argtwo)
   (prog ()
       (when (and (eqcar argone 'aplus2)
                  (integer-constant? argtwo))
           (when  (integer-constant? (cadr argone))
                  (setq argtwo (list 'quote
                               (wplus2 (constant-value (cadr argone))
                                       (constant-value argtwo))))
                  (setq argone (caddr argone))
                  (go fertig))
           (when  (integer-constant? (caddr argone))
               (setq argtwo (list 'quote
                              (wplus2 (constant-value (caddr argone))
                                      (constant-value argtwo))))
               (setq argone (cadr argone))
               (go fertig))
        )
        (when (and (eqcar argtwo 'aplus2)
                   (integer-constant? argone))
           (when  (integer-constant? (cadr argtwo))
                  (setq argone (caddr argtwo))
                  (setq argtwo (list 'quote
                               (wplus2 (constant-value (cadr argtwo))
                                       (constant-value argone))))
                  (go fertig))
           (when  (integer-constant? (caddr argtwo))
               (setq argone (cadr argtwo))
               (setq argtwo (list 'quote
                              (wplus2 (constant-value (caddr argtwo))
                                      (constant-value argone))))
               (go fertig))
        )
       fertig
        (return (list 'aplus2 argone argtwo))
))
 
(def-pass-1-integer-reform  wplus2 u
  % PA1FN:  (wPLUS2 (inf X) y) -> (aplus x y)
 (prog (a b)
  (setq u (groupop (cons 'wplus2 u)))
  (when (not (eqcar u 'wplus2)) (return u))
  (setq a (cadr u)) (setq b (caddr u))
  (return
     (cond ((or (&infp a) (&infp b))
                  `(aplus2 ,(stripinf a) ,(stripinf b)))
           (t u))
)))
 
(de &infp (a) % find (field xxxx '27 '37) to remove inf !
   (and (eqcar a 'field) (equal (caddr a)  '(quote 27))
                         (equal (cadddr a) '(quote 37))))
 
(de stripinf (a) (if (&infp a) (cadr a) a))
 
(if_system Cray-2
 (progn
 (put 'specarithmask 'opencode '((speccon1 (reg 1))))
 (de arithmetic-workhorse (a b gen wgen)
  (cond ((and (knowntobeinteger a) (knowntobeinteger b))
            `(,wgen ,a ,b))
        ((knowntobeinteger a)
           `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wand
                        (wplus2 (wrshift ($Local $$x$$) (quote 18))
                               (quote 1))
                        (specarithmask))
                     (quote 1))(,wgen ($Local $$y$$) ($Local $$x$$) ))
                    ((quote t) (,gen ($Local $$y$$) ($Local $$x$$)))))
              ,b ,a))
        ((knowntobeinteger b)
           `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wand
                        (wplus2 (wrshift ($Local $$x$$) (quote 18))
                                (quote 1))
                        (specarithmask))
                       (quote 1)) (,wgen ($Local $$x$$) ($Local $$y$$) ))
                      ((quote t)  (,gen  ($Local $$x$$) ($Local $$y$$) ))))
              ,a ,b))
        (t `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wor (wand
                             (wplus2 (wrshift ($Local $$x$$) (quote 18))
                                     (quote 1))
                             (specarithmask))
                           (wand
                             (wplus2 (wrshift ($Local $$y$$) (quote 18))
                                     (quote 1))
                             (specarithmask)))
                       (quote 1))(,wgen ($Local $$x$$)($Local $$y$$)))
                      ((quote t) (,gen ($Local $$x$$) ($Local $$y$$)))))
             ,a ,b))
   )  )) 
 (de arithmetic-workhorse (a b gen wgen)
  (cond ((and (knowntobeinteger a) (knowntobeinteger b))
            `(,wgen ,a ,b))
        ((knowntobeinteger a)
           `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wand
                        (wplus2 (wrshift ($Local $$x$$) (quote 18))
                               (quote 1))
                        (quote 8#1777777777777777))
                     (quote 1))(,wgen ($Local $$y$$) ($Local $$x$$) ))
                    ((quote t) (,gen ($Local $$y$$) ($Local $$x$$)))))
              ,b ,a))
        ((knowntobeinteger b)
           `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wand
                        (wplus2 (wrshift ($Local $$x$$) (quote 18))
                                (quote 1))
                        (quote 8#1777777777777777))
                       (quote 1)) (,wgen ($Local $$x$$) ($Local $$y$$) ))
                      ((quote t)  (,gen  ($Local $$x$$) ($Local $$y$$) ))))
              ,a ,b))
        (t `((lambda (($Local $$x$$) ($Local $$y$$))
              (cond ((wleq
                      (wor (wand
                             (wplus2 (wrshift ($Local $$x$$) (quote 18))
                                     (quote 1))
                             (quote 8#1777777777777777))
                           (wand
                             (wplus2 (wrshift ($Local $$y$$) (quote 18))
                                     (quote 1))
                             (quote 8#1777777777777777)))
                       (quote 1))(,wgen ($Local $$x$$)($Local $$y$$)))
                      ((quote t) (,gen ($Local $$x$$) ($Local $$y$$)))))
             ,a ,b))
)     )) 
 
(def-pass-1-integer-reform   plus2 (a b)
        (arithmetic-workhorse a b 'plus2 'wplus2))
 
(def-pass-1-integer-reform   difference (a b)
        (arithmetic-workhorse a b 'difference 'wdifference))
 
(def-pass-1-integer-reform   times2 (a b)
        (arithmetic-workhorse a b 'times2 'longtimes2))
 
(def-pass-1-integer-reform   quotient (a b)
        (arithmetic-workhorse a b 'quotient 'longquotient))
 
(def-pass-1-integer-reform   add1 (a)
        (arithmetic-workhorse a '(quote 1) 'plus2 'wplus2))
 
(def-pass-1-integer-reform   sub1 (a)
        (arithmetic-workhorse a '(quote 1) 'difference 'wdifference))
 
(def-pass-1-integer-reform   minus (a)
        (arithmetic-workhorse '(quote 0) a 'difference 'wdifference))
 
(def-pass-1-integer-reform   onep (a)
        (arithmetic-workhorse '(quote 1) a
         '(lambda (($Local &&x&&) ($local &&y&&))
                         (eqn ($local &&y&&) (quote 1.0))) 'eq))
 
(def-pass-1-reform   zerop (a)
        (arithmetic-workhorse '(quote 0) a
         '(lambda (($Local &&x&&) ($local &&y&&))
                         (eqn ($local &&y&&) (quote 0.0))) 'eq))
 
(def-pass-1-reform   minusp (a)
        (arithmetic-workhorse a '(quote 0) 'minusp 'wlessp))
 
(def-pass-1-reform   greaterp(a b)
        (arithmetic-workhorse a b 'greaterp 'wgreaterp))
 
(def-pass-1-reform   lessp (a b)
        (arithmetic-workhorse a b 'lessp 'wlessp))
 
% fast-functions
 
(def-pass-1-integer-reform   fast-plus (a b)
        (arithmetic-workhorse a b 'plus2 'wplus2))
 
(def-pass-1-integer-reform   fast-difference (a b)
        (arithmetic-workhorse a b 'difference 'wdifference))
 
(def-pass-1-integer-reform  fast-times (a b)
         (arithmetic-workhorse a b 'times2 'longtimes2))
 
(def-pass-1-integer-reform  fast-quotient (a b)
        (arithmetic-workhorse a b 'quotient 'longquotient))
 
(def-pass-1-integer-reform   remainder (a b)
        (arithmetic-workhorse a b 'remainder
                  '(lambda (($Local $$z$$) ($Local $$zz$$))
                            (longdivide-remainder
                             (longdivide ($Local $$z$$)
                                         ($Local $$zz$$))))))
 
(def-pass-1-integer-reform   divide (a b)
        (arithmetic-workhorse a b 'divide
                    '(lambda (($Local $$z$$) ($Local $$zz$$))
                            (progn
                             (longdivide ($Local $$z$$) ($Local $$zz$$))
                             (cons (longdivide-quotient)
                                   (longdivide-remainder))))))
 
(def-pass-1-reform   equal (a b)
  % PA1FN: convert EQUAL to EQ if possible
  (cond
   ((or (eval-eq-able? a)     (eval-eq-able? b))      `(eq ,a ,b))
   ((or (numeric-constant? a) (numeric-constant? b))  `(eqn ,a ,b))
   ((and (knowntobeinteger a) (knowntobeinteger b))
            `(eq ,a ,b))
   ((or (knowntobeinteger a) (knowntobeinteger b))
            `(eqn ,a ,b))
   ))
 
(def-pass-1-reform   eqn (a b)
  (cond
   ((and (knowntobeinteger a) (knowntobeinteger b))
            `(eq ,a ,b))
   ))
 
(de knowntobeinteger (i)
     (when (and (pairp i) (memq (car i) '($FLUID $LOCAL $GLOBAL )))
            (setq i (cadr i)))
      (cond ((numberp i) (intp i))
            ((eqcar i 'setq) (knowntobeinteger (caddr i)))
            ((eqcar i 'progn) (knowntobeinteger (lastcar i)))
            ((flagp i 'INTEGER) T)
            ((eq (cdr (atsoc i cray-datatypes*) ) 'INTEGER) T)
            ((not (pairp i)) nil)
            ((flagp (car i) 'INTEGER) T)
            ((get (car i) 'returns-untagged-value?) T)
            ((flagp (car i) 'INTEGER-TRANSPARENT)
                 (knowntobeintegeraux (cdr i)))
)    )
 
(de knowntobeintegeraux (l)
      (cond ((null l) t)
            ((knowntobeinteger (car l))
                           (knowntobeintegeraux (cdr l)))
            (t nil)
)    )
 
(flag '(abs max2 min2) 'INTEGER-TRANSPARENT)
(flag '(length linelength ceiling fix inf tag) 'INTEGER)
 
(fluid '(cray-datatypes*))
 
(def-pass-1-function prog (u vbls)
  % Pick up PROG vars, ignore labels.
 (let ((cray-datatypes* cray-datatypes*))
  (setq vbls (append (second u) vbls))
  (cons 'prog
      (cons (pass-1-list (second u) vbls)
            (pass-1-prog-body (cddr u) vbls)
 )          )))
 
(dm declare11 (m)
  (let ((u (cadr m)) (v (caddr m)))
      (setq cray-datatypes* (cons (cons u v) cray-datatypes*))
      nil))
 
(def-pass-1-reform   getmem (addr)
 
  % Convert GETMEM into MEMORY expression: (MEMORY var const),
  % where var is the variable part & const is the constant part (tagged, of
  % course).  ADDR has been through pass 1, which does constant
  % folding & puts any constant term at the top level, if possible.
 
  (when (&infp addr) (setq addr (cadr addr)))
 
  (if (eqcar addr 'loc)
    (second addr) % GETMEM LOC x == x
    (cons 'memory
      (cond ((or (eqcar addr 'wplus2) (eqcar addr 'aplus2))
              (if (constant? (third addr))
               (&uninf (rest addr))
               (&uninf (list (third addr) (second addr)))))
            ((and (eqcar addr 'wdifference) (constant? (second addr)))
               (list (list 'wminus (third addr)) (second addr)))
            (t (list addr '(wconst 0)))
            ))))
 
(de &uninf (u)
  (cond ((&infp (car u)) (cons (cadar u) (cdr u)))
        (t u)))
 
(de pa1r-and (l)
    (if (or (pass-1-test-ands l) (eqcar (lastcar l) '$local)) l
                                    % value must be returned
          (binaryandp (cdr l))))
 
 
(de binaryAndp (l)
   (prog (res ll it)
      (setq ll l)
                  % the first has to be done anyway
      (setq res (list (binaryAndOperandp (car ll) nil)))
      (setq ll (cdr ll))
 loop
      (when (null ll) (go ready))
      (setq it (catch 'binary-and-or (binaryAndOperandp (car ll) t) ))
      (when (not it) (go ready))
      (setq res (nconc res (cons it nil)))
      (setq ll (cdr ll))
      (go loop)
 ready
      (when (wlessp (length res) 2) (return `(AND . ,l)))  % do nothing
      (setq res (mappand res))
      (setq res `(EQ (quote 0) ,res))
      (if (null ll) (return res))
      (if (null (cdr ll))
         (return `(AND ,res ,(car ll))))
      (return `(AND ,res ,(binaryAndp ll)))
    ))
 
(fluid '(tagtest*))
 
(setq tagtest*
    '( (pairp . (WXOR (WRSHIFT x (quote 59)) (wconst pair-tag)))
       (floatp . (WXOR (WRSHIFT x (quote 59)) (wconst floatnum-tag)))
       (atom  . (WRSHIFT (WLDZERO (WXOR (WRSHIFT x (quote 59))
                                        (wconst pair-tag)))(quote 6)) )
       (intp .
               ((lambda(($Local $$u$$))
                   (wrshift
                     (Wand (Wdifference ($Local $$u$$) (wconst negint-tag))
         %                 (Wdifference (wconst posint-tag) ($Local $$u$$)))
                           (Wminus ($Local $$u$$)))
                    63) % is negative if not intp
                )
                (WRSHIFT x (QUOTE 59))
       )       )
       (fixp .
               ((lambda(($Local $$u$$))
                   (wrshift
                     (Wand (Wdifference ($Local $$u$$) (wconst negint-tag))
                           (Wdifference (wconst bignum-tag) ($Local $$u$$)))
                    63) % is negative if not fixp
                )
                (WRSHIFT x (QUOTE 59))
       )       )
       (numberp .
               ((lambda(($Local $$u$$))
                   (wrshift
                     (Wand (Wdifference ($Local $$u$$) (wconst negint-tag))
                        (Wdifference (wconst floatnum-tag) ($Local $$u$$)))
                    63) % is negative if not numberp
                )
                (WRSHIFT x (QUOTE 59))
       )       )
    ))
 
(de binaryAndOperandp (l hard)
   (prog(test carl val)
      (when (null l) (return nil))
      (when (numberp l)(return l))
      (when (atom l)(go no))
      (setq carl (car l))
      (setq val
        (cond
         ((or
            (eq carl '$LOCAL)
            (eq carl 'QUOTE) )
            (binaryAndOperandp
                `(NOT (EQ ,l (QUOTE NIL))) hard))
         ((and
          (eq carl 'EQ)
          (or
            (not hard)
            (and
              (islocalref (cadr l))
              (islocalref (caddr l)) )))
    %         (memq (caadr l)  '($Local QUOTE))
    %         (memq (caaddr l) '($Local QUOTE)) )))
          `(WXOR ,(cadr l) ,(caddr l)) )
         ((eq carl 'NE)
          (binaryAndOperandp `(NOT (EQ . ,(cdr l))) hard))
         ((eq carl 'NOT)
          `(WRSHIFT (WLDZERO ,(binaryAndOperandp (cadr l) hard)) (quote 6)))
         ((eq carl 'AND)
           (binaryandp (cdr l)))
         ((and (setq test(atsoc carl tagtest*))
               (eqcar (cadr l) '$LOCAL))
           (subst (cadr l) 'x (cdr test)) )
       ))
       (when val (return val))
  no   (when hard
            (throw 'binary-and-or nil))
       (return
          `(WRSHIFT (WLDZERO (WXOR (QUOTE NIL) ,l))(quote 6)) )
   ))
 
(de islocalref (l) % test, if only local quantities are involved
    (or
       (memq (car l) '($Local QUOTE))
       (and
          (eqcar (car l) 'LAMBDA)
          (equal (cadar l) '(($Local $$$$$$U)))
          (eqcar (cadr l) 'WRSHIFT)
          (islocalref (cadadr l)) )))
 
 
(de mappand (l)
   (cond 
     ((null (cdr l)) (car l))
     (t `(WOR ,(car l) ,(mappand (cdr l)) )) ))
 
(put 'and 'pass-1-reform 'pa1r-and)
 
(de pa1m-intp (l)
    `(wleq (wand
                (wplus2 (wrshift ,(cadr l) 59) 1)
                8#37)
          ,(plus2 posint-tag 1)))
 
(put 'intp 'pass-1-macro 'pa1m-intp)
 
(de pa1m-fixp (l)
    `(wleq (wand
                (wplus2 (wrshift ,(cadr l) 59) 1)
                8#37)
          ,(plus2 bignum-tag 1)))
 
(put 'fixp 'pass-1-macro 'pa1m-fixp)
 
(de pa1m-numberp(l)
    `(wleq (wand
                (wplus2 (wrshift ,(cadr l) 59) 1)
                8#37)
          ,(plus2 floatnum-tag 1)))
 
(put 'numberp 'pass-1-macro 'pa1m-numberp)
 
(de pa1m-eqcar (l)
    `(eq 0 (eqcar-intern ,(cadr l) ,(caddr l))))
             % eqcar-intern is in OPENCODED-FNS
 
(put 'eqcar 'pass-1-macro 'pa1m-eqcar)
 
(de pa1m-and (l)
    (if (pass-1-test-ands (cdr l))
       `(and . ,(pass-1-flatten-ands (cdr l)))
        nil))
 
(de pass-1-test-ands (l)
     (cond ((atom l) nil)
           ((eqcar (car l) 'and) t)
           (t (pass-1-test-ands (cdr l))) ))
 
(de pass-1-flatten-ands (l)
     (cond ((atom l) l)
           ((eqcar (car l) 'and) (append (cdar l) (cdr l)))
           (t (cons (car l) (pass-1-flatten-ands (cdr l))))))
 
(put 'and 'pass-1-macro 'pa1m-and)
 
(de pa1m-digit (u) `((lambda (x) (eq 1 (wrshift
                      (wand (wdifference 8#057 (inf x))
                            (wdifference (inf x) 8#072))
                                 63))) ,(cadr u)))
 
(put 'digit 'pass-1-macro 'pa1m-digit)
 
(de pa1m-liter (u)
  `((lambda (&u& &infu&)
    (setq &infu& (inf &u&))
     (eq 0 (wor (wxor (tag &u&) id-tag)
               (wrshift
                  (wand (wor (wdifference &infu& 8#141) % a
                             (wdifference 8#172 &infu&)) % z
                        (wor (wdifference &infu& 8#101) % A
                             (wdifference 8#132 &infu&)) % Z
                   ) 63)))) ,(cadr u) 0))
 
(put 'liter 'pass-1-macro 'pa1m-liter)
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Stuff for PASS2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(de &memory (u)
   (or (eqcar u 'memory) (eqcar u '$fluid) (eqcar u '$global)))
