       % Pass-2 selected from
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PC:PASS-2-AND-3.SL
% Description:  Passes two and three of the PSL compiler
% Author:       M. Griss, E. Benson, J. Peterson
% Created:      since 4-5-81
% Modified:     14-Nov-84 14:54:33 (Brian Beach)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%      Based on original Standard LISP Compiler by Griss and Hearn
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 1-Sep-88 Winfried Neun, ZIB Berlin
%  Introduced "never-come-back" functions, that are called without
%  the expectation that they will return. Typically errfunctions that
%  throw at their end. (flag never-return)
%  That allows better code for the calling functions. (&call)
% 31-Dec-84 10:00 (Brian Beach)
%  Merged COMPILER-RUNNING fix and COMCOND fix.
% 25-Jul-84 14:40:00 (Mike Creech)
%  Corrected comments for *COMPILER-RUNNING.
% 13-Jul-84 11:24:59 (Mike Creech)
%  Added fluild *COMPILER-RUNNING in COMPD that it T when the compiler is running
%  and NIL otherwise.
% 14-Nov-84 14:54:19 (Brian Beach)
%  Fixed COMCOND so that (COND (T 1) (T 2)) will work.
% 01-Aug-84 13:33:00 (Martin Griss)
%  Fixed &PROTECTARGREGS.
% 3:25pm  Friday, 30 March 1984  (Nancy Kendzierski)
%  Removed debugging (print args) statement from &comapply-aux.
% 2:37pm  Friday, 30 March 1984  (Brian Beach)
%  Fixed &COMCONS.  (setf exp (rest exp)) was missing.
% 27-Mar-84 08:13:21 (Brian Beach)
%  Added optimization for IDAPPLY and CODEAPPLY similar to that done for APPLY.
%  Added ARGUMENT-COUNT property for special compiler functions so that the
%  count doesn't have to be checked in each COMPFN.
%  Added use of *FAST-APPLY-LOAD Cmacro for loading the machine specific register
%  that hold the function or address being called.
% 20-Mar-84 14:00:20 (Brian Beach)
%  Removed if-system in COMAPPLY by introducing a new Cmacro: *FAST-APPLY-LOAD
%  to load the appropriate register with the function for FAST-APPLY.
% 14 Mar 1984 1340-PST (Nancy Kendzierski)
%  Changed calls on REGP to calls on RegisterP (it's defined in
%  common-predicates.sl, part of pass-1-lap)
% 29-Feb-84 11:20:26 (Brian Beach)
%  Removed &CONSTTAG function (equivalent function now is RESULT-TAG).
%  Changed ASSOCOPV so that it does not reorder non-constant parameters.
%  Removed functions for handling LISPVAR and PUTLISPVAR.
%  Ran through pretty printer.
% 19-Jan-84 18:32:20 (Tim Tillson and Sam Sands)
%   Revised function &COMCONS to change the order of evaluation of arguments.
% 18-Jan-84 (Sam Sands)
%   Added quick hack in &COMAPPLY to get it working on the VAX.  A better
%   solution must be found.
% 16-Jan-84 15:48:07 (Tim Tillson)
%    Added wconstevaluable-for-compiler which evaluates expressions only if
%    the function in question has the property DOFN; taken from the routine
%    in pc/anyreg-cmacro.sl
% 30 Dec 1983 1052-PST (Brian Beach)
%  Added stuff from build file.
% 20-Dec-83 16:09:48 (Sam Sands)
%    Fixed MKWCONST to fetch the actual constant value from the property list
% 19 Dec 1983 0900-PST (Brian Beach)
%   Changed WCONSTP so that any integer that is a fixnum or inum is a
%   wconstant.  This change will go away as soon as syslisp mode goes away.
%   Only relevant on the DEC20 where there are still RED files.
% 14-Dec-83 14:44:06 (Tim Tillson)
%   Incorporated Brian's changes into Sam's D-model version;
%   should eventually remove all references to SYSLISP
%
%  10 November, 1983  Brian Beach
%    WCONSTP       - Accepts only forms prefixed by "WCONST" and, in
%                    SYSLISP mode, identifiers declared as WCONSTS.
%    MKWCONST      - Removed lookup of WCONST properties, just adds
%                    WCONST prefix now.
%    ISAWCONST     - Removed check of *SYSLISP
%    PASS-1-WARRAY - Added
%    PASS-1-NOOP   - Added
%    &ASSOCOPV     - Now uses WconstEvaluable for constants.
%    &GETNUM       - Now calls WconstEvaluable to evaluate constants.
%
%  Nov-83  Sam Sands
%   Revised to support D REGISTER model, fix bugs, some optimizations, etc.
%
%     10/31/83 Gerrie Shults      Converted to Lisp syntax
%
% Edit by Cris Perdue, 20 May 1983 1004-PDT
% Added Galway's patch to !&ARGLOC, also extra check and warn clause
% Edit by Cris Perdue, 9 May 1983 1619-PDT
% Removed imports of compl-extra -- now an "extra" module that may be loaded
% with the compiler at a user's discretion.
% Edit by Cris Perdue, 28 Apr 1983 1758-PDT
% Now imports compl-extra, LISP syntax extras
%  <PSL.COMP>COMPILER.RED.21, 30-Mar-83 08:19:44, Edit by KESSLER
%  Fix !&REGMEM to safely check cadr.
% <PSL.COMP>COMPILER.RED.21,  4-Mar-83 01:55:41, Edit by GALWAY
%   Modified CMACRO expansion to check for correct number of arguments in
%   the function call being expanded.
% MLG: 15 Dec
%   added additional arguments to
%    Compiler BUG message in &LOCATE to get more info
%  <PSL.COMP>COMPILER.RED.19,  3-Dec-82 18:21:21, Edit by PERDUE
%  Removed REFORMNE, which was over-optimizing sometimes
%  <PSL.COMP>COMPILER.RED.18,  1-Dec-82 15:59:45, Edit by BENSON
%  Fixed car of atom bug in &PaApply
%  <PSL.COMP>COMPILER.RED.4, 20-Sep-82 11:40:31, Edit by BENSON
%  Slight improvement to "FOO not compiled" messages
%  <PSL.COMP>COMPILER.RED.2, 20-Sep-82 10:32:51, Edit by BENSON
%  (DE FOO (LIST) (LIST LIST)) does the right thing
%  <PSL.COMP>COMPILER.RED.10, 10-Sep-82 12:43:27, Edit by BENSON
%  NONLOCALSYS calls NONLOCALLISP if not WVAR or WARRAY
%  <PSL.COMP>COMPILER.RED.9, 10-Sep-82 09:53:08, Edit by BENSON
%  Changed error and warning messages
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(compiletime
(setq list-of-replaced-modules
    (list  'Pa1r-wplus2 'Pa1r-plus2 'pa1r-difference 'pa1r-onep
           'Pa1r-zerop  'Pa1r-minusp 'pa1r-greaterp  'pa1r-lessp
           'Pa1r-fast-plus  'Pa1r-fast-difference
           'Pa1r-fast-quotient 'pa1r-equal
           'Pa1f-prog 'pa1r-getmem
           '&comidapply '&comaddressapply '&comapply 'consed-list
           '&comcase '&comandor '&pass3)))
 
(compiletime (flag list-of-replaced-modules 'lose))
 
(global '(erfg*
        maxnargs&
        globalgensym&     % list of symbols to be re-used by the compiler
        *nolinke
        *ord
        *r2i
        *unsafebinder
        *noframefluid
        *useregfluid
        *installdestroy
        *usingdestroy
        *showdest
        *msg
        ))
 
(fluid '(dfprint*
       *plap
       *syslisp
       alsts&
       flagg&
       name&
       golist&
       codelist&
       branchpoint&       %   REPLACES CONDTAIL!&   scs
       tail&              %                         scs
       llngth&
       narg&              %% A-list database ((register . info) . . . ).  The info
                          %% for each register seems to be a list of "expressions"
                          %% whose values are available in the register. /csp
       regs&
       exitt&
       lblist&
       jmplist&
       slst&              %% A-list database ((($LOCAL <name>) (FRAME <number>)) . . . )
                            %% Maps local name to frame slot and vice-versa. /csp
       stomap&
       lastactualreg&
       switch&
       toplab&
       freebound&
       status&
       regs1&
       pregs&
       gpregs& %  scs
       destreg&
       exitregs&
       dest&
       environment&
       holemap&
       localgensym&      % traveling pointer into GLOBALGENSYM!&
       ))
 
 
%COMMENT machine dependent parts are in a separate file;
%
%COMMENT these include the macros described below and, in addition,
%      an auxiliary function !&MKFUNC which is required to pass
%      functional arguments (input as FUNCTION <func>) to the
%      loader. In most cases, !&MKFUNC may be defined as MKQUOTE;
%
%COMMENT Registers used:
%1-MAXNARGS!&      used for args of link. result returned in reg 1;
%
%COMMENT Macros used in this compiler;
%
%COMMENT The following macros must NOT change REGS!& 1-MAXNARGS!&:
%!*ALLOC nw            allocate new stack frame of nw words
%!*DEALLOC nw            deallocate above frame
%!*ENTRY      name type noargs   entry point to function name of type type
%                     with noargs args
%!*EXIT                  EXIT to previously saved return address
%!*JUMP adr              unconditional jump
%!*LBL adr            define label
%!*LAMBIND regs alst      bind free lambda vars in alst currently in regs
%!*PROGBIND alst            bind free prog vars in alst
%!*FREERSTR alst            unbind free variables in alst
%!*STORE reg floc      store contents of reg (or NIL) in floc
%
%COMMENT the following macro must only change specific register being
%      loaded:
%
%!*LOAD reg exp            load exp into reg;
%
%COMMENT the following macros do not protect regs 1-MAXNARGS!&:
%
%!*LINK fn type nargs        link to fn of type type with nargs args
%!*LINKE fn type nargs nw  link to fn of type type with nargs args
%                       and EXITT!& removing frame of nw words;
%
%
%COMMENT variable types are:
%
%  LOCAL            allocated on stack and known only locally
%  GLOBAL      accessed via cell (GLOBAL name) known to
%              loader at load time
%  WGLOBAL      accessed via cell (WGLOBAL name) known to
%              loader at load time, SYSLISP
%  FLUID            accessed via cell (FLUID name)
%            known to loader. This cell is rebound by LAMBIND/
%            PROGBIND if variable used in lambda/prog list
%            and restored by FREERSTR;
%
%COMMENT global flags used in this compiler:
%!*UNSAFEBINDER      for Don's BAKER problem...GC may be called in
%            Binder, so regs cant be preserved
%!*MODULE      indicates block compilation (a future extension of
%            this compiler)
%!*NOLINKE       if ON inhibits use of !*LINKE macro
%!*ORD            if ON forces left-to-right argument evaluation
%!*PLAP            if ON causes LAP output to be printed
%!*R2I            if ON causes recursion removal where possible;
%
%
%COMMENT global variables used:
%
%DFPRINT!*      name of special definition process (or NIL)
%ERFG!*            used by REDUCE to control error recovery
%MAXNARGS!&      maximum number of arguments permitted in implementation;
%
%
%
%%Standard LISP limit;
%
%COMMENT fluid variables used:
%
%ALSTS      alist of fluid parameters
%FLAGG      used in COMTST, and in FIXREST
%FREEBOUND indicates that some variables were FLUID
%GOLIST      storage map for jump labels
%PREGS   A list of protected registers
%CODELIST  code being built
%BRANCHPOINT   The point (in SLST) of most recent conditional branch     scs
%TAIL          The accumulated SLST on other branches of a conditional   scs
%CONDTAIL                                           (no longer used )    scs
%LLNGTH      cell whose CAR is length of frame
%NAME      NAME!& of function being currently compiled
%FNAME!&      name of function being currently compiled, set by COMPILE
%NARG      number of arguments in function
%REGS      known current contents of registers as an alist with elements
%      of form (<reg> . <contents>)
%EXITT      label for *EXIT jump
%EXITREGS List or register statuses at return point
%LBLIST      list of label words
%JMPLIST      list of locations in CODELIST!& of transfers
%SLST      association list for stores which have not yet been used
%STOMAP      storage map for variables
%SWITCH      boolean expression value flag - keeps track of NULLs;
%
 
% ---- Support Procedures --------
 
(de &mkfunc (fn)
  % Makes a procedure Name into a procedure object for compilation.
  % Later do FUNARGS
  (mkquote fn))
 
(de &anyregp (x)
  % Test if X is an ANYREG
  (flagp x 'anyreg))
 
 (dm localf (u)
   % Declare functions internal, ala Franz
   % Not presently in use.
   (&comperror "De-implemented macro used: LOCALF")
   )
 
%************************************************************
%        The compiler
%************************************************************
 
(de compile (x)
  % Top level compile entry - X is list of functions to compile
  (for (in fname& x)
       (do (let ((exp (getd fname&)))
           (cond ((null exp)
                (&compwarn (list "No definition for" fname&)))
               ((codep (cdr exp))
                (&compwarn (list fname& "already compiled")))
               (t (compd fname& (car exp) (cdr exp)))
               )))))
 
(de COMPD (NAME& TY Exp)
% COMPD - Single function compiler
% NAME!& is either the name of the function or the value
% ***Code**Pointer**, indicating that a code pointer is to be returned.
% Makes sure function type is compilable.  Sets up CFNTYPE
% property if necessary so the function's type will be known
% during its compilation and thereafter.
% Removes the property when through if it is redundant.
% Calls &COMPROC to compile and adds entry point to the LAP code generated.
% Shows LAP code when PLAP is on.
% Runs LAP.
 
% *COMPILER-RUNNING is a fluid variable that CANNOT be setf'ed to T at the
% beginning of this function and to NIL at the end of this function.
% This is because its purpose is to detect recursive calls on the compiler.
% A recursive call with this approach will set the variable to nil when
% it leaves COMPD -- making it nil to the calling COMPD.
 
   (let
       ((*COMPILER-RUNNING T))
 
       (PROG nil
       (cond ((not (FLAGP TY 'COMPILE))
            (progn (&COMPERROR (list "Uncompilable function type" TY))
                   (Return nil))))
       (cond ((not (EqCar Exp 'lambda))
 
            (progn (&COMPERROR (list "Attempt to compile non-lambda expression" Exp))
                   (Return nil))) (t
%/        ELSE IF !*MODULE THEN MODCMP(NAME!&,TY,EXP)
%              ELSE IF DFPRINT!*
%               THEN APPLY(DFPRINT!*,LIST IF TY EQ 'EXPR
%                                  THEN 'DE . (NAME!& . CDR EXP)
%                                 ELSE IF TY EQ 'FEXPR
%                                  THEN 'DF . (NAME!& . CDR EXP)
%                                 ELSE IF TY EQ 'MACRO
%%                                  THEN 'DM . (NAME!& . CDR EXP)
%                                 ELSE IF TY EQ 'NEXPR
%                                  THEN 'DN . (NAME!& . CDR EXP)
%                                 ELSE LIST('PUTD,MKQUOTE NAME!&,
%                                           MKQUOTE TY,
%                                           MKQUOTE EXP))
                               (PROG (X)
                               (cond ((memq TY '(expr fexpr))
                                    (PUT NAME& 'CFNTYPE (list TY))))
                               (Setq X
 
                                 (CONS (list '*ENTRY NAME& TY (Length (Cadr Exp))) (&COMPROC Exp
                                                                               (cond ((memq TY '(expr fexpr))
                                                                                    NAME&)))))
                      (cond (*PLAP (foreach Y in X do (PRINTF "%f%p%n" Y))))
                  % ***Code**Pointer** is a magic token that tells
                  % COMPD to return a code pointer instead of an ID
                  (cond ((equal NAME& '***Code**Pointer***)
                       (Setq NAME& (LAP X))) (t
 
                                        (progn (LAP X)
                                             %this is the hook to the assembler. LAP must
                                             %remove old function definition if it exists;
                                             (cond
                                              ((and (Setq X (GET NAME& 'CFNTYPE)) (EqCar (GETD NAME&) (car X)))
                                               (REMPROP NAME& 'CFNTYPE)))))))))
 
       (Return NAME&))))
 
(de &comproc (exp name&)
  % COMPROC does the dirty work - initializes variables and gets the
  % three passes going.
  % compiles a function body, returning the generated LAP;
  (prog (narg& codelist& flagg& jmplist& lblist& localgensym& llngth&
               regs& regs1& alsts& exitt& toplab& slst& stomap& condtail&
               freebound& holemap& pregs& switch& exitregs& rn)
        (setq localgensym& globalgensym&)
        (setq pregs& nil)
        (setq regs& nil)
        (setq llngth& 0)
        (unless (eqcar exp 'lambda)
          (&comperror (list "Attempt to compile a non-lambda expression"
                       exp))
          (return nil))
        (setq narg& (length (cadr exp)))
        (setq exitregs& nil)
        (setq exitt& (&genlbl))
        (setq toplab& (&genlbl))
        (setq stomap& nil)
        (setq codelist& (list '(*alloc (*framesize))))
        (&attlbl toplab&)
        (setq exp (compiler-pass-1 exp))
        (when (greaterp narg& maxnargs&)
          (&comperror (list "Too many arguments" narg&)))
        (setq alsts& (&varbind (cadr exp) t))
        % Generate LAMBIND
        (setq rn 1)
        (for (from i 1 (length (cadr exp)) 1)
             (do (setq regs& (&addrvals (&mkreg i) regs& (list (&ireg i))))))
        (&pass2 (caddr exp))
        (&freerstr alsts& 1)
        %Restores old fluid bindings          %  1?  scs
        (&pass3)
        (when (or *installdestroy *showdest)
          (&installdestroy name&))
        (&reformmacros)
        % Plugs compile time constants into macros. FIXFRM?
        (&remtags)
        % Kludge
        (return codelist&)))
 
%-------------------------------------------------------------------
% Pass2 of the COMPILER
%-------------------------------------------------------------------
(de &installdestroy (name&)
  %         scs
  % determine which (if any) registers are unaltered by the function.
  % Print this information out if !*SHOWDEST, install it on the
  % property list of the function if !*INSTALLDESTOY
  (prog (r usedregs)
        (when (equal (setq usedregs (&regsused codelist& nil nil t)) 'all)
          (return nil))
        % Incoming registers have a distinguished value, IREG n, placed
        % in register n.  If this value remains, it has not been destroyed.
        (for (from i 1 narg& 1)
             (do (when (member (&ireg i) (&regval (setq r (&mkreg i))))
                   (setq usedregs (delete r usedregs)))))
        (when (null usedregs)
          (setq usedregs '((reg 1))))
        (when *installdestroy
          (put name& 'destroys usedregs))
        (when *showdest
          (prin2 name&)
          (prin2 " DESTROYS ")
          (prin2t usedregs))))
 
(de &ireg (n)
  % returns form ('IREG n)
  % Generates initial VALUES for REGISTERs to see if they are changed
  % changed by compialtion. Support for INSTALLDESTROY
  %/ Use of vector for efficiency, or for UNIQUE forms?
  (if (and (greaterp n 0) (leq n 15))
    (getv '[nil (ireg 1) (ireg 2) (ireg 3) (ireg 4) (ireg 5) (ireg 6)
            (ireg 7) (ireg 8) (ireg 9) (ireg 10) (ireg 11) (ireg 12)
            (ireg 13) (ireg 14) (ireg 15)]
          n)
    (list 'ireg n)))
 
 
%************************************************************
%      True Pass 2                                  *
%************************************************************
(de &pass2 (exp)
  % scs
  % Initialize STATUS!&=1  (Top level)
  % 1 means want value
  % 2 means no value necessary
  % 3 means inside PROG, want value from RETURN but not from sequential exp's
  % 4 means inside PROG, no RETURN value necessary
  (&comval exp 1))
 
(de &comval (exp status&)
  % Compile EXP.  Special cases: if STATUS!&>1 (compiling for side effects),
  % anyreg functions are ignored since they have no side effects.
  % Otherwise, top level ANYREG stuff is factored out and done via a LOAD
  % instead of a LINK.
  (cond ((&anyreg exp) (if (greaterp status& 1)
           (progn
             (unless (member (car exp) '(quote $local $fluid))
               (&compwarn
                (list "Value of" exp "not used, therefore not compiled")))
             nil)
           (&lreg1 exp)))
        % Just a LOAD
        % When not all ANYREG
        ((&anyregfnp exp) % Is the top level an ANYREG fn?
         (if (greaterp status& 1)
           (progn
             (&comval (cadr exp) status&)
             (&compwarn (list "Top level" (car exp) "in" exp
                         "not used, therefore not compiled"))
             nil)
           (&comexp '(reg 1) exp)))
        % Preserve the anyreg fn
        (t (&comval1 exp stomap& status&)) % no anyregs in sight
      )) 
 
(de &comval1 (exp stomap& status&)
 
  % Generate code which loads the value of EXP into register 1
  %/ examine Patch to COMVAL1 for better register allocation
 
  (cond ((or (&anyreg exp) (&openfnp exp) (&anyregfnp exp))
         (cond ((and (greaterp status& 1) (&nosideeffectp exp)) %   scs
                (&compwarn (list exp " not compiled")))
               (t (&comexp (if (greaterp status& 1)
                     (&alloctemp exp)
                     '(reg 1)) exp))))
 
        ((not (atom (car exp))) % Non atomic function?
         (if (eq (caar exp) 'lambda)
           (&comply (car exp) (cdr exp) status&) % LAMBDA compilation
           (&comperror (list (car exp) "Invalid as function")))) %  Should be noticed in pass 1
 
        % Dispatch built in compiler functions
        ((get (car exp) 'compfn) (&special-comp-function exp status&))
 
        ((eq (car exp) 'lambda)
         (&comperror (list "Invalid use of LAMBDA in COMVAL1" exp)))
 
        (t (&call (car exp) (cdr exp) status&)))  % Call a function
 
  nil)
 
 
(de &alloctemp (exp)
  % Allocate a temporary REG for OPEN exprs.
  % Uses PREGS& to keep track
  % Used only when STATUS!&>1 to
  % set up destination.  Only special case is SETQ.  SETQ tries to put the
  % value of X:=... into a register containing X (keeps variables in the same
  % register if possible.
  (cond ((equal (car exp) 'setq)
         (cond ((equal (car (caddr exp)) 'setq) % Nested setq - move to actual RHS
                (&alloctemp (caddr exp)))
               (t (prog (reg)
                        (cond % LHS variable already in reg?
                              ((and
                                (setq reg
                                 (&rassoc (cadr (cadr exp)) regs&))
                                (not (member (car reg) pregs&)))
                               % and reg must be available
                               (return (car reg)))
                              (t % Return the reg previously used for the var
                               (return (&tempreg))))))))
        (t % Just get a temp
           (&tempreg))))
 
% not SETQ - any old temp will do
(de &call (fn args status&)
  % Issues instructions for Standard CALL on procedure, compiles
  % arguments to egisters 1..N
  (prog (dest& argno)
        (setq argno (length args))
        (if (&anyregp fn)
          (&comexp '(reg 1) (cons fn args))
          (progn
            (&comload args)
            (if (flagp fn 'never-return)
              (progn
                (&attach (list '*jcall-never-return fn argno))
                (return NIL))
              (&attach (list '*link fn (&cfntype fn) argno)))
            (&remmrefs)
            (&remvrefs)
            % Default - all registers destroyed
            (when *usingdestroy
              (setq dest& (get fn 'destroys)))
            (if (null dest&)
              (setq regs& nil)
              (prog (temp)
                    (setq temp nil)
                    (foreach r in regs& do (unless (member (car r) dest&)
                       (setq temp (cons r temp))))
                    (setq regs& temp)))))))
 
(de &comlis (exp)
  % COMLIS  returns LIST of "locations" of compiled arguments
  (reversip (&comlis1 exp)))
 
(de &comlis1 (exp)
  % Support for COMLIS, returns reversed list of compiled arguments;
  % Gathers a set of ANYREG expressions denoting
  % the params.  Code for non ANYREG stuff is emitted by ATTACH.  ACUSED is
  % name of psuedo variable holding results of non anyreg stuff.
  %/ See <griss> code to fix an ORD bug. Need to check SIDE-EFFECTS
  %/ more stringently
  (prog (acused y)
    (setq y nil)
    (while exp
      (cond ((or (&constp (car exp))
             (and (&openp (car exp))
                      (or (not *ord) (&nosideeffectpl (cdr exp)))))
           (setq y (cons (car exp) y)))
          % Anyreg stuff is handled later.  Anyreg args are not loaded until after
          % all others.
          % If !*ORD is true, order is still switched unless no side effects
          (t %/  Special coding for top level ANYREG
           (cond ((and
                 (or (&anyregfnp (car exp)) (&openfnp (car exp)))
                 (or (not *ord) (&nosideeffectpl (cdr exp))))
                (setq y
                  (cons (cons (caar exp) (&comlis (cdar exp))) y)))
               % Emit code to place arg in R1, generate a name for the result to put in R1
               (t (&comval (car exp) 1) %    scs
                  (setq acused (list '$local (&gensym)))
                  (setq regs&
                  (&addrvals '(reg 1) regs& (list acused)))
                  % REGS!& the new variable name goes on the code list (rest already emitted)
                  (setq y (cons acused y))))
           (&saver1) %                scs
           ))        % place arg in memory while doing others
      (setq exp (cdr exp)))
    (return y)))
 
(de &saver1 ()
  % Marks contents of REG 1 for storage.
  % Will allocate a FRAME location for ($LOCAL ) if needed
  (prog (x)
        (setq x (&regval '(reg 1)))
        % Contents of R1
        (cond ((or (null x) (not (&varp (car x)))) (return nil))
              % Dont save constants
              ((not (assoc (car x) stomap&)) (&frame (car x))))
        % For temporaries
        % as generated in COMLIS
        (&storelocal (car x) '(reg 1))))
 
% Emit a store
(de &comply (fn args status&)
  % COMPFN: Compile LAMBDA expression.
  %/ Exists a bug and inefficiency. See <griss> files that
  %/ fix this.
  % SHouldnt just LOADARGs, should behave as some sort of PROG
  (prog (oldstomap oldcode alsts& vars n i)
        %     scs
        (setq vars (cadr fn))
        % Compile args to the lambda
        (setq n (length args))
        (when (greaterp n maxnargs&)
          (&comperror (list "Too many arguments in LAMBDA form" fn)))
        % Put the args into registers
        (&comload args)
        %  scs
        % Enter new ENVIRONMENT!&
        (setq args (&remvarl vars))
        % The stores that were protected;
        (setq oldstomap stomap&)
        %     scs
        (setq oldcode codelist&)
        %     scs
        (setq i 1)
        % Put this junk on the frame
        (setq alsts& (&varbind vars t))
        %Old fluid values saved;
        % compile the body
        (&comval (caddr fn) status&)
        % Restore old fluids
        (&freerstr alsts& status&)
        % Go back to the old ENVIRONMENT!&
        (&fixfrm oldstomap oldcode)
        %  let's see this work      scs
        (&rstvarl vars args)))
 
(de &loadargs (args status& pregs&)
  % Load a sequence of expressions into the registers
  % for CALL, COMPLY etc. Use after call on COMLIS1,
  % args are "ANYREG/OPEN" expressions in reverse order
  (prog (fn destreg& n)
        (setq n (length args))
        (when (greaterp n maxnargs&)
          (&comperror (list "Too many arguments" args)))
        (while args
          % Generate a load for each arg
          (progn
            (setq destreg& (&mkreg n))
            (&loadopenexp destreg& (car args) status& pregs&)
            (setq pregs& (cons destreg& pregs&))
            (setq n (difference n 1))
            (setq args (cdr args))))))
 
(de &comload (args)
  (prog (oldstomap oldcode)
        (setq oldstomap stomap&)
        (setq oldcode codelist&)
        (&loadargs (&comlis1 args) 1 pregs&)
        %Emits loads to registers
        (&cleartemps oldstomap oldcode)))
 
(de &comexp (reg exp)
  (prog (oldstomap oldcode)
        (setq oldstomap stomap&)
        (setq oldcode codelist&)
        (&loadopenexp reg (cons (car exp) (&comlis (cdr exp))) status&
                      pregs&)
        (&cleartemps oldstomap oldcode)))
 
(de &cleartemps (oldstomap oldcode)
  (when (neq oldstomap stomap&)
    (while (ne stomap& oldstomap)
      (&clrstr (caar stomap&))
      (setq stomap& (cdr stomap&)))
    (&fixfrm oldstomap oldcode)))
 
(de &lreg1 (x)
  % Load register 1 with X
  (&loadopenexp '(reg 1) x 1 pregs&))
 
% 1?    scs
(de &loadopenexp (destreg& arg status& pregs&)
  % FORCES LOADOPENEXP1 TO LOAD INTO DESTREG                           scs
  % (cond ((or (&ANYREG ARG) (&RAssoc ARG Regs&)) (&LREG DESTREG (&LOCATE ARG)))
  %       (t (PROG (DEST)
  %          (Setq DEST (&LOADOPENEXP1 DESTREG ARG STATUS PREGS PREGS))
  %          (cond ((LESSP STATUS 2) (&LREG DESTREG DEST)))))
  %       )
  (prog (dest)
        (setq dest (&argloc arg))
        (when (lessp status& 2)
          (&lreg destreg& (&locate dest)))))
 
(de &loadopenexp1 (destreg& arg status& gpregs& pregs&)
  % RETURNS DESTREG!&, WHICH MAY BE MODIFIED BY CALLOPEN IF STATUS>0   scs
  % Generates code to load an OPEN/or ANYREG expression into
  % the target register. Uses OPEN and ANYREG tables
  %/ Some bug relating to reusing a register before issuing store,
  %/ lose in &LOCATE. See <griss>partial fixes
  (prog (r)
        (cond %   ((&ANYREGFNP ARG)                                    I think this is dead
              %    (progn (&LOADOPENEXP DESTREG& (Cadr ARG) 1 PRegs&)       %  1? scs
              %       (&LREG DESTREG& (&LOCATE (CONS (car ARG) (CONS DESTREG& (cddr ARG)))))))
              (t %  Must be an open function
                 (prog (opfn adjfn anyregargs)
                       (setq anyregargs (&remopen destreg& (cdr arg) 1))
                       (setq opfn (get (car arg) 'openfn))
                       (if (idp opfn)
                         (apply opfn (list destreg& anyregargs arg))
                         (setq destreg&
                          (&callopen opfn destreg& anyregargs (car arg)))))))
        (return destreg&)))
 
(de &remopen (destreg& args status&)
  %? Returns a LIST of ARGUMENT locations (ARGLOCs)
  (foreach arg in args collect (&argloc arg)))
 
(de &argloc (arg)
  %  RETURNS A DESTREG!& MODIFIED BY LOADOPENEXP                      scs
  %? Returns a STORE or LHS form for argument
  %? Also marks REG in use on PREGS for safety in TempAlloc
  (prog (loc dstreg)
        (when (and (eqcar arg 'quote)       % according discussion with
                   (intp (cadr arg))        % A.C.Hearn, J. Padget
                   (leq (abs (cadr arg)) bitsperword)) % R. Kessler
                    (return arg))           % of 20th August 1988
 
        (cond ((eqcar arg '$name) %         (setq ARG (cadr ARG))     % see if this works            scs
               (cond ((eqcar (cadr arg) '$local) (return arg))
                     %   yuck
                     (t (setq arg (cadr arg)))))
              (t (when (setq loc (&rassoc arg regs&))
                   (setq pregs& (cons (car loc) pregs&))
                   (return (car loc)))))
        (when (&constp arg)
          (return arg))
        %  scs
        (when (&anyreg arg)
          (return arg))
        %   (cond ((&ANYREGFNP ARG) (Return (CONS (car ARG) (CONS (&ARGLOC (Cadr ARG))
        %                                            (cddr ARG))))))
        (when (or (null destreg&) (member destreg& pregs&))
          (setq destreg& (&tempreg)))
        %    (cond ((FLAGP (car ARG) 'MEMMOD)
        %       (progn (&LOADOPENEXP DESTREG& ARG 2 PRegs&)       % could be better    scs
        %         (Return
        %         (cond ((eq (car (Cadr ARG)) 'MEMORY)       % Galway
        %            (Cadr ARG))
        %             (t
        %            (progn (cond ((neq (car (Cadr ARG)) '$NAME)       % /csp
        %                        (&CompWarn (list "Arg to !&ARGLOC `" ARG "'unrecognized!"))))
        %                   (Cadr (Cadr ARG)))))))))
        (setq dstreg (&loadopenexp1 destreg& arg status& pregs& pregs&))
        %   scs
        (if (registerp dstreg)
          (setq pregs& (cons dstreg pregs&))
          (&protectargregs (cdr dstreg)))
        (return dstreg)))
 
(de &protectargregs (exp)
  % PROTECT ALL THE REGISTERS IN EXPRESSON    scs
  (when exp
    (let ((e (car exp)))
      (cond ((registerp e)
           (setq pregs& (cons e pregs&)))
          ((null (pairp e)) nil)    %/MLG
          ((eq (car e) 'quote) nil) %/MLG
          (t (&protectargregs (cdr e)))) %/MLG
      (&protectargregs (cdr exp))
      )))
 
(de &callopen (opfn dest& args op)
  %  RETURNS DEST!&, WHICH MAY BE MODIFIED BY MACRO PATTERN           scs
  %% OPFN is e.g. (ASSOCPAT *WPLUS2)
  %% DEST& (fluid) is e.g. (reg 1), a destination location
  %% ARGS is e.g. ((reg 1) (reg 2)), specifying locations of args
  %% OP is e.g. WPLUS2, apparently the LISP operator name. /csp
  %? Basic OPENFN pattern matcher
  %? Issues code to CALL expression that compiles open.
  %? Uses Pattern Tables to check register usage.
  %? Does Local REGISTER allocation
  %? Setup new ENVIRONMENT for generated temps
  (prog (pats params adjfn regfn environment&)
        (setq pats (car opfn))
        % E.g. ASSOCPAT /csp
        (when (idp pats)
          (setq pats (get pats 'pattern)))
        (setq params (cons op (cdr opfn)))
        % I.e. (<lisp fn> <cmacro name>) /csp
        (setq adjfn (car pats))
        (setq regfn (cadr pats))
        % I.e. ('*SET ... ) ('*DO ... ), etc. /csp
        (when adjfn
          (setq args (apply adjfn (list args))))
        (setq pats (cddr pats))
        % ((<pattern> . <cmacros) . . . ) /csp
        (while (and (not (null pats)) (not (&matches (caar pats) args)))
          (setq pats (cdr pats)))
        (when (null pats)
          (&comperror (list "Compiler bug - no pattern for" (cons op args)))
          (return nil))
        %% (CAR PATS) matched present situation.  Substitute first arg for A1,
        %% second for A2, CMACRO name for MAC, etc. in CMACROS in pattern. /csp
        (foreach mac in (cdar pats) do
                 (&emitmac (&subargs mac args params)))
        %% Generally used to fix up REGS& database to reflect what has happened as
        %% part of "execution" of this operation. /csp
        (when regfn
          (if (idp regfn)
            (apply regfn (list op args))
            (&emitmac (&subargs regfn args params))))
        (return dest&)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Basic CALLOPEN Pattern Matching and Subsitution functions
%%% &MATCHES is the pattern matcher for the pattern - cmacro mapping tables
%%% used by CALLOPEN.
(de &matches (pat subj)
  % Recursive matcher.
  %? Cases handled are: ?
  (cond ((eqcar pat 'quote) (equal (cadr pat) subj))
        ((null pat) (null subj))
        ((eqcar pat 'noval)
         (and (greaterp status& 1) (&matches (cdr pat) subj)))
        ((atom pat) (apply (get pat 'matchfn) (list subj)))
        (t (and (pairp pat)
                (cond ((pairp subj) %   scs
                       (and (&matches (car pat) (car subj))
                        (&matches (cdr pat) (cdr subj))))
                      (t (and (&matches (car pat) nil)
                          (&matches (cdr pat) nil))))))))
 
(de &any (u)
  t)
 
% Always TRUE
(de &dest (u)
  % Checks if U is the DEST& register
  (equal u dest&))
 
(de &usesdest (u)
  % Indicates an anyreg which uses DEST!& at any level
  % for Reservation of REGS?
  (or (&dest u) (and (pairp u) (&usesdestl (cdr u)))))
 
(de &usesdestl (u)
  % Support for USESDEST
  (and (pairp u)
       (or (&dest (car u)) (&usesdest (car u)) (&usesdestl (cdr u)))))
 
(de &regfp (u)
  (or %? Check if U is a REG or LOCAL
      (&regp u) (eqcar u '$local)))
 
(de &regn (u)
  (or %? Check if U is REG or LOCAL or NIL
      (&regp u) (eqcar u '$local) (equal u ''nil)))
 
(de &mem (u)
  %? Check if can store into U
  (and (not (or (equal u ''nil) (eqcar u '$local)))
       (or (&constp u) (&varp u) (equal (car u) 'memory))))
 
(de &notanyreg (u)
  %? Check if CANT use as an anyreg
  (or (&mem u) (&regfp u)))
 
(de &regnp (r)
  (and (registerp r) (not (member r gpregs&))))
 
(de &typeregnp (r regkindp)
  (and (apply regkindp (list r)) (&regnp r)))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Pattern subsitution functions used in CALLOPEN /csp
(de &subargs (mac args params)
  % Substitute into each element of MAC /csp
  (foreach arg in mac collect (&subarg arg args params)))
 
(de &subarg (arg args params)
  % Basic substituter.  ARG is substituted into.  ARGS is list of expressions
  % for actual operands for CMACROs as passed to &CALLOPEN.  PARAMS is list of
  % LISP and CMACRO operator names, respectively, to substitute for FN and
  % MAC. /csp
  (prog (argfn)
        (return (cond ((eqcar arg 'quote) (cadr arg))
                      ((pairp arg) (&subargs arg args params))
                      ((equal arg 'dest) dest&)
                      ((setq argfn (get arg 'substfn))
                       (apply argfn (list arg args params)))
                      (t (&comperror
                          (list "Compiler bug" arg "invalid in macro")))))))
 
%% Popular (all?) SUBSTFNs for use by &SUBARG.  Trivial except for ones that
%% get temps. /csp
(de &arg1 (arg args params)
  % LOCATE ARG1 for OPEN funs
  (&locate (car args)))
 
(de &arg2 (arg args params)
  % LOCATE ARG2 for OPEN funs
  (&locate (cadr args)))
 
(de &arg3 (arg args params)
  % LOCATE ARG3 for OPEN funs
  (&locate (caddr args)))
 
(de &arg4 (arg args params)
  % LOCATE ARG4 for OPEN funs
  (&locate (cadddr args)))
 
(de &param1 (arg args params)
  % Reference PARAM1 for OPEN funs
  (car params))
 
(de &param2 (arg args params)
  % Reference PARAM2 for OPEN funs
  (cadr params))
 
(de &negparam2 (arg args params)
  % Reference the NEGATIVE JUMP for PARAM2 of OPEN tests
  (get (cadr params) 'negjmp))
 
(de &param3 (arg args params)
  % Reference PARAM3 for OPEN funs
  (caddr params))
 
(de &param4 (arg args params)
  % Reference PARAM4 for OPEN funs
  (cadddr params))
 
(de &gettemp (tname args params)
  %  scs 
  % Allocate a new TEMPREG, and store name/value in ENVIRONMENT
  (&gettypetemp tname args params '&any))
 
(de &gettypetemp (tname args params regkindp)
  %  scs 
  (prog (tn)
        (return (if (setq tn (assoc tname environment&))
                  (cdr tn)
                  (progn
                    (setq tn (&temptypereg regkindp))
                    (setq environment& (cons (cons tname tn) environment&))
                    (setq pregs& (cons tn pregs&))
                    tn)))))
 
(de &gettemplbl (lname args params)
  % Allocate a new TEMPLBL, and store name/value in ENVIRONMENT
  (prog (lab)
        (return (if (setq lab (assoc lname environment&))
                  (cdr lab)
                  (progn
                    (setq lab (&genlbl))
                    (setq environment&
                     (cons (cons lname lab) environment&))
                    lab)))))
 
%%% End of subsituter.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(de &gensym ()
  % gensym local to compiler, reuses symbols
  % on LOCALGENSYM list
  (prog (symb)
        (when (null (cdr localgensym&))
          (rplacd localgensym& (list (gensym))))
        (setq symb (car localgensym&))
        (setq localgensym& (cdr localgensym&))
        (return symb)))
 
(de &comperror (u)
  % General COMPILER Error message
  %? Maybe should CALL THROW, rather than just setting ERFG!*
  %? Or maybe ERFG!* collects errors for end?
  (progn 
    (errorprintf "***** in %P: %L" name& u)
    (setq erfg* t)))
 
(de &compwarn (u)
  % General Compiler WARNING message
  % not as error
  (and *msg (errorprintf "*** in %P: %L" name& u)))
 
(de &emitmac (mac)
  %% Seems to serve 2 quite different purposes.  For ordinary CMACROS, just
  %% attaches its argument to the codelist, using &ATTACH.
  %% For *SET, etc., updates REGS& database to reflect new value in some
  %% register. /csp
  % Support function for CALLOPEN,
  % After Pattern match, emit form
  % clear destroyed regs, etc.
  (prog (emitfn)
        (cond ((equal (car mac) '*do) (apply (cadr mac) (cddr mac)))
              ((equal (car mac) '*destroy) %% Register now contains nothing of interest /csp
               (foreach reg in (cdr mac) do
                        (setq regs& (delasc reg regs&))))
              ((equal (car mac) '*set) %% Compute all "expressions" whose value one might want to consider as
               %% being in the register, enter them them into database with register.
               %% See defn. of &REMREGSL for details. /csp
               (setq regs&
                     (&repasc (cadr mac) (&remregsl (caddr mac)) regs&)))
              ((setq emitfn (get (car mac) 'emitfn))
               (apply emitfn (list mac)))
              (t (&attach mac)))))
 
(de &emitload (m)
  % Open support, emit a LOAD
  (&lreg (cadr m) (caddr m)))
 
(de &emitstore (m)
  % Open support, EMIT a STORE
  (&storevar (caddr m) (cadr m)))
 
(de &emitjump (m)
  % Open support, emit a jump
  (&attjmp (cadr m)))
 
(de &emitlbl (m)
  % Open support, emit a label
  (&attlbl (cadr m)))
 
(de &emitmemmod (m)
  % open support, emit a memory modifying instruction
  % will update SLST and REGS& since changed state
  (prog (y x)
        (setq x (cadr m))
        (&remrefs x)
        (when (eqcar x '$local)
          (&refstore x))
        %   scs   assumes $NAME
        (when (eqcar x '$local)
          (setq m (cons (car m) (cons (&getfrm x) (cddr m)))))
        (&attach (cons (get (car m) 'unmemmod) (cdr m)))))
 
(de &uses (m)
  % PROTECTS A REGISTER WHICH WILL BE NEEDED LATER                 scs
  (setq pregs& (cons (cadr m) pregs&)))
 
(de &return (m)
  % INDICATES WHERE THE VALUE OF THE OPEN CODED FUNCTION IS        scs
  (setq dest& (cadr m)))
 
(de &noanyreg (args)
  % Support to patterns - register adjustment functions
  % remove all ANYREG stuff except top level MEMORY
  (if (null args)
    nil
    (cons (&noanyreg1 (car args)) (&noanyreg (cdr args)))))
 
(de &noanyreg1 (arg)
  % Support to NOANYREG
  (if (and (&anyregfnp arg) (not (eqcar arg 'memory)))
    (&loadtempreg arg)
    arg))
 
(de &inreg (args)
  %? Check if in regs; ensure that loaded?
  (if (not (&regfp (car args)))
    (list (&loadtempreg (car args)))
    args))
 
(de &regmem (args)
  %? Check if in REG or MEM/ensure that is loaded?
  (progn 
    (setq args (&noanyreg args))
    (if (and (&mem (car args)) (pairp (cdr args)) (&mem (cadr args)))
      (cons (&loadtempreg (car args)) (cdr args))
      args)))
 
(de &destmem (args)
  % A1 in DEST!&, A2 in MEM, rest (if any) not anyreg
  (progn 
    (setq args (cons (car args) (&noanyreg (cdr args))))
    (cond ((greaterp status& 1) (if (&regfp (car args))
             args
             (cons (&loadtempreg (car args)) (cdr args))))
          ((or (&dest (cadr args)) (&usesdest (cadr args)))
           (&destmem (cons (car args)
                      (cons (&loadtempreg (cadr args)) (cddr args)))))
          ((neq (car args) dest&) (&lreg dest& (&locate (car args)))
           (cons dest& (cdr args)))
          (t args))))
 
(de &destmema (args)
  % put either a1or A2 into DEST!&, the other to MEM.
  (cond ((equal (car args) dest&) % A1 = DEST!&, make A1 mem or reg
         (if (and (&notanyreg (cadr args)) (not (&usesdest (cadr args))))
           args
           (&loadtemp2 args)))
        ((equal (cadr args) dest&) % A2 = DEST!&, make A2 mem or reg
         (if (and (&notanyreg (car args)) (not (&usesdest (car args))))
           args
           (&loadtemp1 args)))
        ((or (&notanyreg (cadr args)) (not (&notanyreg (car args)))) % A2 is MEM or A1 is anyreg: make A1 the destination
         (progn
           (when (or (not (&notanyreg (cadr args)))
                     (&usesdest (cadr args)))
             (setq args (&loadtemp2 args)))
           (&lreg dest& (&locate (car args)))
           (cons dest& (cdr args))))
        (t % Make A2 the DEST!& - only when A2 is anyreg and a1 is mem
           (progn
             (when (or (not (&notanyreg (car args)))
                       (&usesdest (car args)))
               (setq args (&loadtemp1 args)))
             (&lreg dest& (&locate (cadr args)))
             (list (car args) dest&)))))
 
(de &loadtemp1 (u)
  % Bring first arg into a temp
  (cons (&loadtempreg (car u)) (cdr u)))
 
(de &loadtemp2 (u)
  % put second arg in a temp
  (cons (car u) (cons (&loadtempreg (cadr u)) (cddr u))))
 
(de &consargs (args)
  %? Reassemble arguments, may need temp
  (if (or (and (not (&anyregfnp (cadr args))) (neq (cadr args) dest&))
          (and (not (&anyregfnp (car args))) (neq (car args) dest&)))
    args 
    (list (car args) (&loadtempreg (cadr args)))))
 
(de &loadtempreg (arg)
  % Load ARG into a temporary register.  Return the register.
  (prog (temp)
        (setq temp (&tempreg))
        (setq pregs& (cons temp pregs&))
        (&lreg temp (&locate arg))
        (return temp)))
 
(de &fixregtest (op args)
  % Fixes up the registers after a conditional jump has been emitted.
  % For JUMPEQ and JUMPNE, equalities can be assumed in REGS!& or REGS1!&
  % For other jumps, REGS!& copied onto REGS1!&.
  (&fixregtest1 op (first args) (second args)))
 
(de &fixregtest1 (op a1 a2)
  % Support for FIXREGTEST
  (progn 
    (setq regs1& regs&)
    (when (or (equal op 'eq) (equal op 'ne))
      (if (not (&regp a1))
        (progn
          (when (&regp a2)
            (&fixregtest1 op a2 a1)))
        (progn
          (if (equal op 'eq)
            (setq regs1& (&addrvals a1 regs1& (&remregs a2)))
            (setq regs& (&addrvals a1 regs& (&remregs a2)))))))))
 
(de &setregs1 (op args)
  % Saves REGS& in REGS!& for COND branches
  (setq regs1& regs&))
 
(de &locate (arg)
  %   scs
  (cond ((eqcar arg '$name) (cadr arg))
        %   yuck
        ((eqcar arg '$local) (&refstore arg) %  scs
         (cadr (assoc arg stomap&)))
        (t arg)))
 
(commentoutcode (de &locate (x)
                  %  scs
                  % Find the location of a variable
                  % ie. an ANYREG expression that can be directly loaded
                  % Checks for: Const, $NAME, MEMORY,REG,ANYREG, FRAME
                  %             updates SLST if LOCAl (ie, "needed")
                  (prog (y vtype)
                        % Constants are their own location
                        (when (or (atom x) (eqcar x 'label) (&constp x))
                          (return x))
                        (when (eqcar x '$name)
                          (return (cadr x)))
                        (when (equal (car x) 'memory)
                          (return
                           (cons (car x)
                            (cons (&locate (cadr x)) (&locate (cddr x))))))
                        %  scs
                        (when (setq y (&rassoc x regs&))
                          (return (car y)))
                        % If in a register, return the register number
                        % Registers are their own location
                        % For ANYREG stuff, locate each constant
                        (when (&anyregfnp x)
                          (return (cons (car x) (&locatel (cdr x)))))
                        (unless (eqcar x '$local)
                          (return x))
                        % Since the value of the variable has been referenced, a previous store was
                        % justified, so it can be removed from SLST!&
                        % Must be in the frame, otherwise make nonlocal (really ought to be an error)
                        % Frame location (<=0) is returned
                        (&refstore x)
                        %  scs
                        (when (setq y (assoc x stomap&))
                          (return (cadr y)))
                        % Nasty compiler bug.  Until we fix it, tell the user to simplify expressions
                        (&comperror
                         (list
                          "Compiler bug: expression too complicated, please simplify"
                          x))
                        (&comperror
                         (list " Length CodeLIST " (length codelist&)))
                        (return ''verycomplexexpression)))
                % just so it doesn't blow up and can be seen
                (de &locatel (u)
                  % Locate a LIST of values
                  (foreach x in u collect (&locate x))))
 
(de &sameval (reg val actualval)
  (or (equal reg val)
      (prog (regval)
            (when (setq regval (&regval reg))
              (return (while actualval
                        (when (member (car actualval) regval)
                          (return 't))
                        (setq actualval (cdr actualval))))))))
 
(de &lreg (reg val)
  % PARTIAL BUG FIX, &REMREGS RETURNS A LIST SO MUST TAKE CAR OF ACTUALVAL scs
  % Load register REG with value U. V (always NIL except when called from
  % LOADARGS) is a list of other loads to be done.
  % Adjust REGS&, issue MOVE if needed
  (prog (actualval)
        (when (&sameval reg val (setq actualval (&remregs val)))
          (return nil))
        (&attach (list '*move val reg))
        (setq regs& (&repasc reg actualval regs&))))
 
(de &jumpt (lab)
  % Emit a JUMPT
  (&attach (list '*jumpnoteq lab '(reg 1) ''nil)))
 
(de &jumpnil (lab)
  % EMit a JUMPN
  (&attach (list '*jumpeq lab '(reg 1) ''nil)))
 
%---------- Functions for Handling Non-local Variable Binding ----
(de &varbind (vars lambp)
  % bind FLUID variables in lambda or prog lists;
  % LAMBP is true for LAMBDA, false for PROG;
  (prog (i vlocs vnames fregs y reg tail)
        (setq i 1)
        (foreach x in vars do (progn
                   (setq reg (&mkreg i))
                   (cond ((eqcar x '$global) % whoops
                          (progn
                            (&compwarn
                             (list "Illegal to bind global" (cadr x)
                              "but binding anyway"))
                            (rplaca x '$fluid))))
                   % cheat a little
                   (cond ((eqcar x '$fluid) (setq freebound& t)
                          (setq vnames (cons x vnames)) (unless *noframefluid
                            (setq vlocs (cons (&frame x) vlocs)))
                          (setq fregs (cons reg fregs)))
                         ((eqcar x '$local) (&frame x)
                          (&storelocal x (if lambp
                             reg
                             nil)))
                         (t
                          (&comperror
                           (list "Cannot bind non-local variable" x))))
                   (when lambp
                     (if (eqcar x '$local)
                       (setq regs& (&repasc reg (list x) regs&))
                       (setq regs& (&repasc reg nil regs&))))
                   (setq i (plus i 1))))
        (when (null vnames)
          (return nil))
        (setq vnames (cons 'nonlocalvars vnames))
        (setq fregs (cons 'registers fregs))
        (setq vlocs (cons 'frames vlocs))
        (setq tail (if *noframefluid
                (list vnames)
                (list vnames vlocs)))
        (if lambp
          (&attach (cons '*lambind (cons fregs tail)))
          (&attach (cons '*progbind tail)))
        (when *unsafebinder
          (setq regs& nil))
        (return tail)))
 
(de &freerstr (alsts& status&)
  % restores FLUID variables  % why STATUS& ?  scs
  (when alsts&
    (&attach (cons '*freerstr alsts&))
    (when *unsafebinder
      (setq regs& nil))))
 
(de &attach (u)
  % ATTACH is used to emit code
  (setq codelist& (cons u codelist&)))
 
(de &storelocal (u reg)
  % marks expression U in register REG for storage
  % setting SLST and removing  unneeded stores from SLST
  (prog (x)
        (when (null reg)
          (setq reg ''nil))
        (setq x (list '*move reg (&getfrm u)))
        % Update list of stores done so far
        (&attach x)
        % Zap out earlier stores if there were never picked up
        % ie, if you store to X, then a ref to X will remove this store from
        % SLST!&.  Otherwise, the previous store will be removed by CLRSTR
        % SLST!& is for variables only (anything else?)
        (&clrstr u)
        (setq slst& (cons (cons u codelist&) slst&))))
 
(de &clrstr (var)
  % removes unneeded stores remembered on SLST
  % and removes the MOVE etc from CODELIST
  (prog (x sl)
        (setq sl slst&)
        (while sl
          (cond ((equal (car (setq x (car sl))) var) (rplaca x nil) % SLST!& := DelQIP(X,SLST!&);  scs
                 (&delmac (cdr x)))
                ((equal (car x) 'branch) (return nil)))
          (setq sl (cdr sl)))))
 
(de &refstore (var)
  (prog (x sl)
        (setq sl slst&)
        (while sl
          (when (equal (car (setq x (car sl))) var)
            (rplaca x nil))
          %   SLST!& := DELETIP(Y,SLST!&);         scs
          (setq sl (cdr sl)))))
 
%----------------- Functions for general tests --------------------
(de &comtst (exp labl)
  % compiles boolean expression EXP.
  % If EXP has the same value as SWITCH!& then branch to LABL,
  % otherwise fall through;
  % REGS& are active registers for fall through,
  % REGS1& for branch;
 
  (prog (x fn reg)
 
    % First factor out NOT's to set up the SWITCH!&
    (while (and (eqcar exp 'eq) (equal (caddr exp) ''nil))
      (setq switch& (not switch&))
      (setq exp (cadr exp)))
 
    % SWITCH!& is assumed to be true by fn's with
    % a flip test
    (when (and (not switch&) (setq fn (get (car exp) 'fliptst)))
      (setq exp (cons fn (cdr exp))))
 
    % Dispatch a built in compiling function
    (if (setq fn (get (car exp) 'opentst))
      (if (atom fn)
      (apply fn (list exp labl))
      (&comopentst fn exp labl pregs&))
      (if (eqcar exp 'quote)
      (progn
        (setq regs1& regs&)
        (when (or (and switch& (cadr exp))
                (and (not switch&) (not (cadr exp))))
          (&attjmp labl))
        )
      (&comtst (list 'ne exp ''nil) labl)))))
 
(de &comopentst (pat exp destlab pregs&)
  %? Pattern match entru for OPENTST functions
  (prog (oldstomap oldcode)
        (setq oldstomap stomap&)
        (setq oldcode codelist&)
        (&callopen pat destlab (&remopen (&tempreg) (&comlis (cdr exp)) 1)
                   (car exp))
        (&cleartemps oldstomap oldcode)))
 
(de &remvarl (vars)
  % Remove variables to avoid name conflicts:  Hide variable names which match
  % new names when entering an inner function.  Other names will be available
  % as global info.  VARS is the list of new variable names, the result is a
  % list of protected stores, ie (var . slst-entry)
  %? What about REGS contents
  % Support for nested environments and branches; ie
  % name of VAR used
  (foreach x in vars collect (&protect x)))
 
(de &protect (u)
  % WON'T HANDLE MULTIPLE INSTANCES
  % Support for REMVARL
  % Delete all references to U from SLST!&
  % return the protected store, to be put back later
  (prog (x)
        (&remrefs u)
        %  scs
        (return (when (setq x (assoc u slst&))
                  (rplaca x nil)
                  (cons u x)))))
 
%  scs
(de &rstvarl (vars lst)
  % Restore a previous set of names.  VARS is the list of variables taken out
  % of the SLST. LST is the list of protected stores.  One or zero
  % stores for each variable.
  %/ What of REGS
  (while vars
    (&rstvar (car vars) (car lst))
    (setq vars (cdr vars))
    (setq lst (cdr lst))))
 
(de &rstvar (var val)
  % Support for RSTVARL
  % Restore a particular variable and STORE
  (prog nil
        (&remrefs var)
        % Remove new occurance from REGS
        (&clrstr var)
        % Flush from SLST, removing unneeded Stores
        % in cas
        % Put back old deferred stores  SLSTS
        (&unprotect val)))
 
(de &unprotect (val)
  %restores VAL to SLST!&;
  (when val
    (rplaca (cdr val) (car val))))
 
%             scs
(de &storevar (u v)
  % The store generated by a SETQ
  (prog (vtype x)
        (&remrefs u)
        (cond ((equal (car u) '$local) % assumes $NAME
               (&storelocal u v))
              (t (&attach (list '*move v u))))
        (when (&regp v)
          (setq regs& (&addrvals v regs& (list (&remregs u)))))))
 
%  scs
(de &references (exp var)
  % True if expression EXP (probably ANYREG) references VAR.
  (or (equal exp var) (if (or (atom exp) (flagp (car exp) 'terminal))
        nil
        (&referencesl (cdr exp) var))))
 
(de &referencesl (exp var)
  % Support for REFERENCES
  (if (null exp)
    nil
    (or (&references (car exp) var) (&referencesl (cdr exp) var))))
 
(de &cfntype (fn)
  % Look up the type of a function using its CFNTYPE property if
  % any.  Return EXPR by default.
  (prog (x)
        (return (cond ((setq x (get fn 'cfntype)) (car x))
                      ((setq x (getd fn)) (car x))
                      (t 'expr)))))
 
(de &genlbl ()
  % General an internal name for a temp label
  (prog (l)
        (setq l (list 'label (&gensym)))
        (setq lblist& (cons (list l) lblist&))
        (return l)))
 
(de &getlbl (labl)
  % Look up the Internal name of label in a prog
  (prog (x)
        (setq x (assoc labl golist&))
        (when (null x)
          (&comperror (list "Compiler bug: missing label" labl)))
        (return (cdr x))))
 
(de &attlbl (lbl)
  % Attach a label to codelist.
  % Handle the (!*LBL x) (!*LBL y) case by making y alias for x
  (if (eq (caar codelist&) '*lbl)
    (&defeqlbl lbl (cadr (car codelist&)))
    (&attach (list '*lbl lbl))))
 
(de &attjmp (lbl)
  % Attach a jump to codelist.
  % handle (JUMP l) (LBL l) case
  (prog nil
        (when (eq (caar codelist&) '*lbl)
          (&defeqlbl lbl (cadr (car codelist&)))
          (&delmac codelist&))
        (when (&transferp codelist&)
          (return nil))
        (&attach (list '*jump lbl))))
 
(de &transferp (x)
  % Check if previously emited CMACRO is an unconditional
  % transfer (jump or call).
  (if (equal (caar x) '*noop)
    (&transferp (cdr x))
    (flagp (if (eq (caar x) '*link)
             (cadar x)
             (caar x)) 'transfer)))
 
(de &defeqlbl (lab1 lab2)
  % Make labels LAB1 and LAB2 aliases for each other
  (setq lblist& (&defeqlbl1 lblist& lab1 lab2)))
 
(de &defeqlbl1 (labs lab1 lab2)
  % Support for DEFEQLBL
  (cond ((member lab1 (car labs)) (if (member lab2 (car labs))
           labs
           (cons (append (&labclass lab2) (car labs))
                 (&delclass lab2 (cdr labs)))))
        ((member lab2 (car labs))
         (cons (append (&labclass lab1) (car labs))
               (&delclass lab1 (cdr labs))))
        (t (cons (car labs) (&defeqlbl1 (cdr labs) lab1 lab2)))))
 
(de &labclass (lab)
  %? Compute label equivalence classes
  (prog (temp)
        (setq temp lblist&)
        (while (and temp (not (member lab (car temp))))
          (setq temp (cdr temp)))
        (return (if temp
                  (car temp)
                  nil))))
 
(de &delclass (lab labs)
  % Remove a label from an equivalence class
  (if (member lab (car labs))
    (cdr labs)
    (cons (car labs) (&delclass lab (cdr labs)))))
 
(de &lbleq (lab1 lab2)
  % Check if LAB1eq to LAB2 (in its equivalence class)
  (member lab1 (&labclass lab2)))
 
(de &frame (u)
  % allocates space for U in frame;
  %? How about reuse of STOMAP entries?
  %? on exit from LAMBDA/PROG, mark as free?
  (prog (z res)
        (setq z (if (null stomap&)
                1
                (plus 1 (cadr (cadar stomap&)))))
        (setq res (&mkframe z))
        (setq stomap& (cons (list u res) stomap&))
        (return res)))
 
(de &getfrm (u)
  % GETFRM returns the frame location of a local variable
  (prog (x)
        (when (setq x (assoc u stomap&))
          (return (cadr x)))
        (&comperror (list "Compiler bug: lost variable" u))))
 
%*************************************************************************
% The following functions determine classes or properties of expressions *
%*************************************************************************
(de &anyreg (u)
  % !&ANYREG determines if U is an ANYREG expression
  %
  % ANYREG expressions are those expressions which may be loaded into any
  % register without the use of (visable) temporary registers.  It is assumed
  % that ANYREG expressions have no side effects.
  %
  % ANYREG expressions are defined as constants, variables, and ANYREG functions
  % whose arguments are ANYREG expressions.  Note that ANYREG functions are
  % not necessarily a part of ANYREG expressions; their arguments may not be
  % ANYREG expressions.
  (or (&constp u) (&varp u) (and (&anyregfnp u) (&anyregl (cdr u)))))
 
(de &anyregl (u)
  (or % Are all expression in list U Anyreg
      (null u) (and (&anyreg (car u)) (&anyregl (cdr u)))))
 
(de &anyregfnp (u)
  % !&ANYREGFNP is true when U is an ANYREG function.  The arguments are not
  % checked
  (&anyregp (car u)))
 
(de &openp (u)
  % does U open compile?
  (or (&constp u) (&varp u)
      (and (or (&anyregfnp u) (&openfnp u)) (&openpl (cdr u)))))
 
(de &openpl (u)
  % Do each expression in list U opencompile
  (or (null u) (and (&openp (car u)) (&openpl (cdr u)))))
 
(de &openfnp (u)
  % true if U is an OPENfn
  (get (car u) 'openfn))
 
(de &constp (u)
  % True if U is a constant expression
  (and (idp (car u)) (flagp (car u) 'const)))
 
(de &varp (u)
  % True if U is a variable: (LOCAL x),(FLUID x), ...
  (and (pairp u) (flagp (car u) 'var)))
 
(de &regp (u)
  % True if U is a Register expressions (REG and EXTRAREG?)
  (and (pairp u) (flagp (car u) 'reg)))
 
(de &nosideeffectp (u)
  % True if the expression U has no side effects.  ANYREG expressions and
  % functions are assumed to have no side effects; other functions must be
  % flagged NOSIDEEFFECT.  All arguments to a function must also be NOSIDEEFFECT.
  (or (&anyreg u)
      (and (or (&anyregfnp u) (flagp (car u) 'nosideeffect))
           (&nosideeffectpl (cdr u)))))
 
(de &nosideeffectpl (u)
  % Support for LIST of NOSIDEEFFECTS
  (or (null u) (and (&nosideeffectp (car u)) (&nosideeffectpl (cdr u)))))
 
