% REDUCE Interface for the PVM (Parallel Virtual Machine)
%
% Winfried Neun, August 1993
%
% The Interface provides an 'algebraic' access to PVM (V 3) 
% (i.e. using the PSL module pvm).


module redpvm;

load pvm;

fluid '(!*all!-procs!* rrhilf);

rrhilf := gtwarray 2;

symbolic procedure REMOTE_kill(p);
  if (p = -1) then 
	 << for each pp in !*all!-procs!* do pvm_kill pp;
	    !*all!-procs!* := nil>>
  else << delq(p,!*all!-procs!*); pvm_kill p>>;


symbolic procedure REMOTE_PROCESS(filename);
    begin scalar info;
	info := start_slaves(1,"reduce");
	if not (filename = "" or null filename) then
		order_slaves(cdr info,1,list('in1,filename,nil),0,0);

	!*all!-procs!* := wgetv (cdr info,0) . !*all!-procs!*;
	return wgetv (cdr info,0);
    end;

symbolic procedure REMOTE_PROCESSES(anz,filename);
    begin scalar info,lilo;
        info := start_slaves(anz,"reduce");
        if not (filename = "" or null filename) then
                order_slaves(cdr info,1,list('in1,filename,nil),0,0);

	for II:=1:anz  do
        	lilo := wgetv (cdr info,ii-1) . lilo;
        !*all!-procs!* := append (lilo,!*all!-procs!*);
        return ('list . lilo)
    end;

symbolic procedure REMOTE_call(p, fcn, params , result);

	(<<params := for each c in cdr params collect reval c;
	   if p = -1 then p := !*all!-procs!* else p := {p};
	   if p = 0 then << p := {first !*all!-procs!*};
			  !*all!-procs!* := append (cdr !*all!-procs!*,p)
			% wrap around
			 >>;
	   for each pp in p do
	    <<  wputv(rrhilf,0,pp);
		order_slaves(rrhilf,1,{'reval, mkquote (fcn . params)}
				,taskname,result)>>;
	   taskname>>) where taskname = mktaskname(p);

symbolic procedure REMOTE_call!*(p, fcn, params , result);

        (<<if p = -1 then p := !*all!-procs!* else p := {p};
           if p = 0 then << p := {first !*all!-procs!*};
                          !*all!-procs!* := append (cdr !*all!-procs!*,p)
                        % wrap around
                         >>;
           for each pp in p do
            <<  wputv(rrhilf,0,pp);
                order_slaves(rrhilf,1,{'apply,mkquote fcn,mkquote params}
                                ,taskname,result)>>;
           taskname>>) where taskname = mktaskname(p);

fluid '(rrnexttask rrtaskresults);

lisp (rrnexttask := 0);

symbolic procedure mktaskname (p);
	p #+ 1025 #* (rrnexttask:=rrnexttask #+ 1);

symbolic procedure REMOTE_receive(task);
   begin scalar res;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
	
      res := assoc(task,rrtaskresults);
      if res then << rrtaskresults := delasc(task,rrtaskresults);
		     return cdr res >>
	 else << sleep (1); return REMOTE_receive(task)>>;
   end;

symbolic procedure REMOTE_wait();
   begin scalar res;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;

      if rrtaskresults then
	return 'list . for each re in rrtaskresults collect car re
         else << sleep (1); return REMOTE_wait()>>;
      end;

symbolic procedure REMOTE_inquire (task);
   begin scalar res;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
      if res := audition() then rrtaskresults := res . rrtaskresults;
        
      res := assoc(task,rrtaskresults);
      return if res then t else nil;
   end;

symbolic procedure REMOTE_discard(handle);
    % dont know what to do
   4711 ;

symbolic procedure local_yield();
	% dont know what to do
    4711 ;


flag('(REMOTE_kill REMOTE_CALL REMOTE_CALL!* REMOTE_inquire REMOTE_receive
	REMOTE_PROCESSES,
       REMOTE_PROCESS,REMOTE_wait,REMOTE_discard,local_input,local_yield),
     'opfn);

endmodule;
