%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BACKTRACE.SL 
% Title:        Functions for displaying the return stack. 
% Author:       ??? 
% Created:      ??? 
% Modified:     22-Mar-84 09:23:19 (Brian Beach) 
% Package:      Kernel 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 20 January 1989 (Winfried Neun)
%  Adapted backtrace for Sparc (SUN4) stack structure and register windows
% 27-May-87 (Harold Carr & Leigh Stoller)
%  Added fluid declaration of in* and out*.
% 22-Mar-84 09:13:24 (Brian Beach)
%  Fixed constants.  Removed WFORS.
% 01-Dec-83 14:12:23 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime
  (flag '(backtrace1 backtracerange backtracerange1 
	  verbosebacktrace1 backtracerange11)  'internalfunction)

  (put  'get-stack-hack 'opencode '((add  (reg st) 2047 (reg 1))))) 

(fluid '(ignoredinbacktrace* options* interpreterfunctions*))
(fluid '(in* out* *dump))

(setq ignoredinbacktrace*
      (list 'eval 'apply 'fastapply 'codeapply 'codeevalapply 'catch
	    'errorset 'evprogn 'toploop 'breakeval 'bindeval 'break 'main
	    'TOP-LOOP-EVAL-AND-PRINT 'PRE-MAIN))
%%%%%%       'stackcheck))

(setq interpreterfunctions* (list 'cond 'prog 'and 'or 'progn 'setq))

(global '(heapupperbound stacklowerbound))

(de interpbacktrace ()
  (prog (here)
	(setq here (get-stack-hack))
	(printf 
	   "Backtrace, including interpreter functions, from top of stack:%n")
	(return (backtracerange here stacklowerbound 1 nil nil))))

(de backtrace ()
  (prog (here x)
	(setq here (get-stack-hack))
	(printf "Backtrace from top of stack:%n")
	(return (backtracerange here stacklowerbound 0 nil nil))))

%(de backtracerange (starting ending interpflag)
%    (backtracerange1 starting ending interpflag 1))

% Stack start is indicated by a forward/0815 item and stackend by a
% forward/4711 item (rather German numbers indeed)
% We must set the 0815 item before windows are rolled out.

(lap '((*entry backtracerange expr 5)
       (*alloc 0)
       (*move 0815 (reg l0))
       (*mkitem (reg l0) forward-tag)
       (*move 1 (reg 4)) 
       (*link backtracerange1 expr 4)
       (*exit 0)))

% rotating the register windows like a merrygoround at least one rotation

(de backtracerange1 (starting ending interpflag num verbose)
  (if (wlessp num 31)   % 31 is a magic number from SPARC architecture
     (backtracerange1 starting ending interpflag (wplus2 1 num) verbose)
     (backtracerange11 starting ending interpflag verbose)
     (terpri))
  NIL % disables tail recursion 
)

(compiletime
 (put 'half_getmem 'opencode '((ld (displacement (reg 1) 0) (reg 1)))))

(de backtracerange11 (starting ending interpflag verbose)
  (prog (x memo validitems nextframe fwd)
    (while (and (or (not (eq (tag (getmem starting)) forward-tag))
	            (not (eq (inf (getmem starting)) 815))))
	  (setq starting (wplus2 starting 8))
)
	     %  (wminus (wtimes2 addressingunitsperitem stackdirection)))))
   % we found top of stack
  lupe         % loop on stack frames
    (setq nextframe (wplus2 2047 (wgetv starting 14))) % backward link in register frame  
    (setq starting  nextframe)

    (setq validitems (getmem starting))

    (when (not (or (eq (wshift validitems -27) forward-tag)
                   (eq (wshift validitems -59) forward-tag)))
         (when *dump  (print "*** Irregular Stack***")) (go lupe))

    (when (eq (wand validitems 8191) 4711) (return nil)) % all done
    (setq validitems (wand (wshift validitems -10) 1023))

    (for (from i 1 validitems)     % looking for btr items in frame 1 - 12
	 (do (if verbose (verbosebacktrace1 (wgetv starting i) 0)
	      (setq memo (wgetv starting i))
	      (when (weq (tag memo) btr-tag)
	             (backtrace1 (mkid (inf memo)) interpflag))))    )
    (when  (setq x (returnaddressp (wgetv starting 15))) % the o7 in stack
	   (if verbose (verbosebacktrace1 (wgetv starting 15) 0)
	               (backtrace1 x interpflag)))  % holds returnaddress
    
    (go lupe)))

(de verbosebacktrace ()
    (unless (member 'addr2id options*)  (load addr2id))
    (backtracerange 0 0 0 0 T))

(de verbosebacktrace1 (x y) % for one item 
     (cond ((and (codep x) (wgreaterp (inf x) heapupperbound))
	                 (writechar (char tab))
	                 (channelwriteunknownitem out* x)
	                 (terpri))
	   ((weq (tag x) btr-tag)  (printf "       %r%n" (mkid (inf x))))
	   ((setq y (returnaddressp x))
	            (printf "%p -> %p:%n"
	               (code-address-to-symbol (inf x)) y))
           ((and (idp x) (wgreaterp (inf x) maxsymbols)) NIL)
				% always problems with stackaddresses!
	   (t (printf "     %p%n" x))))

(de backtrace1 (item code)
  %
  % Code is 1 if Interpreter functions should be printed, 0 if not.
  %
  (unless (memq item ignoredinbacktrace*)
    (unless (and (equal code 0) (memq item interpreterfunctions*))
      (prin1 item)
      (writechar (char blank)))))

(off fast-integers)
