%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:WFLOAT.SL
% Description:  Interface to foreign floating point functions.             
% Author:       RAM, HP/FSD                                                
% Created:      27-Feb-84                                                  
% Modified:     29-Oct-84 09:06:40 (Vicki O'Day)
% Mode:         Lisp                                                       
% Package:                                                                 
% Status:       Experimental (Do Not Distribute)                           
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.        
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions
%
% 04-Apr-87 (Leigh Stoller)
%  Added support for 68020 IEEE negative zero test. See $pnk/arithmetic.sl.
% 29-Oct-86 (Leigh Stoller)
%  Removed definition of writefloat. It has been returned to write-float.sl.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(**neg-zero** **neg-one** unix_error))

(de *wfloat (x y)
  (uxfloat x y))

(de *fplus2 (x y z)
  (or (eq 0 (uxplus2 x y z)) (fperror 'plus)))

(de *fdifference (x y z)
  (or (eq 0 (uxdifference x y z)) (fperror 'difference )))

(de *ftimes2 (x y z)
  (or (eq 0 (uxtimes2 x y z)) (fperror 'times)))

(de *fquotient (x y z)
  (or (eq 0 (uxquotient x y z)) (fperror 'quotient)))

(de *fgreaterp (x y)
  (uxgreaterp x y 't 'nil))

(de *flessp (x y)
  (uxlessp x y 't 'nil))

(de *wfix (x)
  (let ((w(uxfix x)))
    (if (eq unix_error 0) w (fperror 'fix))))

(de *fassign (x y)
  (uxassign x y))

(de *doubletofloat (x y)
  (uxdoubletofloat x y))

% These two statements must be at the end of the file because times2, and
% thus *ftimes2, must be defined before it is used.
(loadtime (progn
    (setq **neg-one** (minus  -1.0))
    (setq **neg-zero** (times2 **neg-one** 0.0))))

% For DCM                                                                  
(de *floattodouble (x y)
  (uxfloattodouble x y))

(de fperror(m) (error 25 (bldmsg "floating point error in %w" m)))


% support for special functions

(compiletime 
  (dm protected-function(u)
     (setq u (cdr u))
    `(de ,(car u) ,(caddr u)
        (or (eq 0 ,(cons (cadr u)(caddr u)))
            (fperror ',(car u))))))

(protected-function uxsin     uxxsin              (r x))
(protected-function uxcos     uxxcos              (r x))
(protected-function uxtan     uxxtan              (r x))
(protected-function uxasin    uxxasin             (r x))
(protected-function uxacos    uxxacos             (r x))
(protected-function uxatan    uxxatan             (r x))
(protected-function uxsqrt    uxxsqrt             (r x))
(protected-function uxexp     uxxexp              (r x))
(protected-function uxlog     uxxlog              (r x))
(protected-function uxatan2   uxxatan2            (r y x))  
