%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% File:        PXNK:FAST-BINDER.SL
% Title:       Fast binding and unbinding routines in LAP
% Author:      Eric Benson
% Created:     12 July 1981
% Modified:   
% Status:      Experimental
% Mode:        Lisp
% Package:     Kernel
%
% Copyright (c) 1981 Eric Benson
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 31-Aug-88 (Julian Padget)
%  Bnd... declarations changed from fluid to global.
% 20-Jul-86 (Leigh Stoller)
%  Converted from syslisp for use with Bobcat 3.4. Originally came from the
%   Sun Kernel directory on Utah-Yampa. I also made changes to the code so
%   that it was not using Sun specific operations. I based it more on the Hp's
%   Gator 3.4 fast-binder.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(global '(BndStkPtr              % The binding stack pointer
	 BndStkLowerBound	% Bottom of the binding stack
	 BndStkUpperBound))	% Top of the binding stack

% TAG( FastBind )

(lap '((!*Entry FastBind expr 0)			% Bind IDs to values
							% in registers.
      (!*move (fluid BndStkPtr) (Reg t1)) 	   	% Load binding pointer.
      (!*move (Indirect (reg st)) (Reg t0))		% Stack return pointer
							% in t0.
Loop
	(cmpi!.b 255 (Indirect (Reg t0)))		% Test for not an
							% instruction, 68000
							% doesn't use opcode
							% Fxxx,	 16#FF?.
	(beq MoreLeft)					
	(!*move (Reg t1) (fluid BndStkPtr))		% Save new pointer.
        (*move (reg t0) (indirect (reg st)))            % Save new stack ptr
	(!*exit 0)	
MoreLeft
	(!*wplus2 (Reg t1) (wconst 8))  		% Make space for 2 more
							% words on binding stk.
	(!*jumpwlessp (label NoOverflow) (Reg t1) (fluid BndStkUpperBound))
	(!*jcall BStackOverflow)			% If overflow,error.
NoOverflow
	(move!.w (displacement (Reg t0) 0) (reg t4))  	% Reg number in t4.
	(*wand (reg t4) (wconst 255))

	(move!.w (displacement (Reg t0) 2) (Reg t3))	% ID number in t3.
	(!*wand (Reg t3) (wconst 65535))

	(!*jumpon (reg t4) 0 6
		  ((label UseNIL)
		   (label UseR1)
		   (label UseR2)
		   (label UseR3)
		   (label UseR4)
		   (label UseR5)
		   (label UseR6)))

	(*wshift (reg t4) 2)      		% times addressingunitsperitem
	(!*move (memory (reg t4) (fluid Argumentblock)) (reg t4))
                            % Retrieve current value of reg from argumentblock

	(!*jump (label GotRegisterValue))
UseNIL
	(!*move (reg NIL) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR1
	(!*move (reg 1) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR2
	(!*move (reg 2) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR3
	(!*move (reg 3) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR4
	(!*move (reg 4) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR5
	(!*move (reg 5) (reg t4))
	(!*jump (Label GotRegisterValue))
UseR6
	(!*move (reg 6) (reg t4))
	(!*jump (Label GotRegisterValue))

GotRegisterValue
	(move!.l (Reg t3) (displacement (Reg t1) -4))   % Save id number

	(*wshift (reg t3) 2)    		% times addressingunitsperitem
	
	(move!.l (reg t3) (reg t2))
	(!*move (memory (Reg t2) (fluid Symval))
                (indirect (Reg t1)))                    % save current value

	(!*move (reg t4) (memory (Reg t2) (fluid Symval)))
							% store reg value
	(!*wplus2 (reg t0) (wconst 4))           	% move pointer forward
	(!*jump (Label Loop))				% try again
))



% TAG( FastUnBind )

(lap '((!*Entry FastUnBind expr 0)			% Unbind last N 
							% entries in bind stack
%
% FastUnBind is called with JSB, followed by byte containing count to
% unbind.
%
	(!*MOVE (fluid BndStkPtr) (Reg t1)) 		% get binding stack 
							%  pointer in t1.
	(move!.l (Indirect (reg st)) (reg t0))		% address in t0
	(clr!.l (Reg t3))
	(move!.w (Indirect (reg t0)) (reg t3))          % get count in t3
	(*wplus2 (Indirect (reg st)) (wconst 2))        % restore bumped addres
	
Loop
	(!*jumpwgreaterp (Label MoreLeft) (Reg t3) (wconst 0))
					% continue if count is > zero
	(!*move (Reg t1) (fluid BndStkPtr))
					% otherwise store binding stack pointer
	(!*exit 0)			% and return
MoreLeft
	(!*jumpwgeq (Label NoUnderflow) (Reg t1) (Fluid BndStkLowerBound))
					% check underflow
	(!*jcall BStackUnderflow)
NoUnderflow
	(move!.l (displacement (Reg t1) -4) (reg t4)) 	% get ID number in t4

	(*wshift (reg t4) 2) 			% times addressingunitsperitem
	(!*move (indirect (reg t1))
                (memory (reg t4) (fluid SymVal)))       % store value

	(!*wdifference (Reg t1) (wconst 8))
				     % adjust binding stack pointer down by 2
	(!*wplus2 (Reg t3) (wconst -1))			% count down by 1
	(!*jump (Label Loop))				% try again
))


