%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:OTHER-IO.SL
% Title:        Miscellaneous input and output functions
% Author:       Eric Benson
% Created:      28 August 1981
% Modified:     03-Aug-84 09:58:56
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:IO-DECLS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 03-Aug-84 09:58:35 (Jim Ambras/CRC)
%  Modified channelprinc to correct type in channelprin2 call.
% 22-Mar-84 11:40:52 (Brian Beach)
%  Added compiletime load of io-decls.
% 01-Dec-83 14:56:28 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load io-decls))
 
(fluid  '(out*)) % Current output channel
 
(fluid '(*raise **low-case)) % controls case conversion of IDs
 
(de channeleject (c)
  % Skip to top of next output page
% (progn (channelwritechar c (char ff))
%        nil))
  (newpage) nil)
 
% write a formfeed
(de eject ()
  % Skip to top of next output page
  (channeleject out*))
 
(de channellinelength (chn len)
  % Set maximum line length
  (prog (oldlen striplen)
        (setq oldlen (wgetv maxline chn))
        (when len
          (if (and (intp len) (wgeq len 0))
            (setf (wgetv maxline chn) len)
            (stderror (bldmsg '"%r is an invalid line length" len))))
        (return oldlen)))
 
% if Len is NIL, just return current
(de linelength (len)
  % Set maximum line length
  (channellinelength out* len))
 
(de channelposn (chn)
  (wgetv lineposition chn))
 
(de posn ()
  % Number of characters since last EOL
  (channelposn out*))
 
(de channellposn (chn)
  (wgetv pageposition chn))
 
(de lposn ()
  % Number of EOLs since last FF
  (channellposn out*))
 
(compiletime
(ds doraise (x)
  (setq x (wxor
           (wlshift
            (wrshift (wand (wdifference (wplus2 (char (lower a)) -1) x)
                      (wdifference (wplus2 x -1) (char (lower z)))) 63)
            5)
          x)))
)
 
(de channelreadch (chn)
  % Read a single character ID
    (prog (x)
        % for Standard Lisp compatibility
        (setq x (channelreadchar chn))
        % converts when *RAISE
    (if **low-case
        (when (and *raise (wgeq x (char !A)) (wleq x (char !Z)))
          (setq x (wplus2 (char !a) (wdifference x (char !A)))))
        (when (and *raise (wgeq x (char !a)) (wleq x (char !z)))
          (setq x (wplus2 (char !A) (wdifference x (char !a)))))
    )
    (return (mkid x))))

(de readch ()
  % Read a single character ID
  (channelreadch in*))
 
(de channelterpri (chn)
  % Terminate current output line
  (progn (channelwritechar chn (char eol))
         nil))
 
(de terpri ()
  % Terminate current output line
  (channelterpri out*))
 
(de princ (x)
  (prin2 x))
 
(de channelprinc (c x)
  (channelprin2 c x))
 
(commentoutcode de readch ()
  % Read a single character ID from IN*
  (prog (charr inn rai)
        % for Standard Lisp compatibility
        % converts lower to upper when *RAISE
 
  (setq inn in*)
  (setq charr (wgetv unreadbuffer inn))
  (setq rai *raise)
 
  (if (eq inn (wand inn maxchannels)) % IMPURE but fast!
 
     (if (neq charr (char null))
      (setf (wgetv unreadbuffer inn) (char null))
      (setq charr (idapply (wgetv readfunction inn) (list inn)))
     )
     (return (noniochannelerror inn "ChannelReadChar"))
  )
  (when rai (doraise charr))
  (return (mkid charr))))
 
