%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXNK:SYSTEM-EXTRAS.SL                                      
% Title:          MS-DOW specific code for PSL                          
% Author:         Herbert Melenk, Winfried Neun 
% Created:        3 May 1990                                           
% Modified:       
% Package:        Kernel                                                   
%
% (c) Copyright 1990 , Konrad-Zuse-Zentrum , all rights reserved.       
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

% 14-May-1990  (H. Melenk)
%  added function SYSTEM for calling MS-DOS programs,

(fluid '(**windows))

(compiletime (load sys-consts sys-macros doshooks))
(compiletime (load inum fast-vector fast-int))

(compiletime (load include))
(compiletime (include "c:/psl/kernel/dos386/windows/psllcall.h"))

(compiletime (flag '(copy-partial-string find-blank
		     ren del)
		   'internalfunction))

(compiletime
  (ds upper-case(c)
     (if (and (wgeq c (char !a)) (wleq c (char !z)))
	 (wplus2 (char A) (wdifference c (char !a)))
	 c)))

(on fast-integers)

(de quit ()
    (errorprintf "%f%nQuitting")
    (cond
     ((wgreaterp breaklevel* 0) (exit-with-status 1))
     (t (exit-with-status 0))))

(lap '((*entry exit-with-status expr 1)
       (*move ($fluid **windows)(reg 2))
       (cmp 1 (reg 2))
       (jne nowin)
    %   (*move 0 (reg 1))
    %   (*call psll-main)
       (*move (wconst EXIT_WITH_STATUS) (reg 4))
       (*call psll-call)
    nowin
       (*wand (reg 1)(wconst 255))
       (*wplus2 (reg 1) (wconst (times 256 16#4c)))
       (int 16#21) ))

(de exitlisp () (quit))

(de current-directory()
   (when (weq **windows 1) (windows-error))       
   (let (x y z)
     (386asm 
	   (setf (386reg ah) 16#19) 
	   (386INT 16#21) 
	   (setq z (386reg AL))) 
	(setq x (mkstring 67))
	(setf (strbyt (strinf x) 0) (wplus2 (char A) z))
	(setf (strbyt (strinf x) 1) (char :)) 
	(setf (strbyt (strinf x) 2) (char !\)) 
	(386asm
	   (setf (386reg ah) 16#47)
	   (setf (386reg ESI) (wplus2 3 (strbase (strinf x)))) 
	   (386INT 16#21))
	(setq y nil)
	(ifor (from i 0 64 1)
	      (do (when (and (null y)
			     (eq (strbyt (strinf x) i) 0))
			(setq y i))))
	(copy-partial-string x y 0)))

(de copy-partial-string (str l pos)
   % copy a partial string from str (length l) pos
    (while (and (wleq pos l) (eq (strbyt (strinf str) pos) (char blank)))
	   (setq pos (iadd1 pos)))
    (let ((new (mkstring (wdifference l pos))))
	(ifor (from i pos l 1)
	      (do (setf (strbyt (strinf new)(wdifference i pos))
			(strbyt (strinf str) i))))
	new))

(de find-blank(str l pos)
    (while (and (wleq pos l) (not (eq (strbyt (strinf str) pos)(char blank))))
	   (setq pos (iadd1 pos)))
    pos)

%-------------------------------------------------------------------------
%   function file-status
%   command dir/ls
 
(declare-warray filestatus-work size 11)
(compiletime (flag
   '(expand-status-block set-dta get-dta find-first find-next
     ) 'internalfunction))
 
(de filestatus(str do)
   (when (weq **windows 1) (windows-error))       
   (let* ((fn (fnexpand str))
	  (dta (set-dta filestatus-work))
	  (r (find-first (strbase (strinf fn)) )))
      (if (not (izerop r))
	  nil
	  (expand-status-block) )))
 
(de expand-status-block()
  (let ((block filestatus-work) n tm da c)
    (ifor (from i 0 12 1)
	  (do (progn
	       (when (not (izerop c)) (setq c (byte block (wplus2 i 30))))
	       (push (if (izerop c) '!  (wand c 127)) n) )))
     (setq n (cons '!" (append n '(!"))))
   `((mode nil . ,(byte block 21))
     (time nil . ,(setq tm (wand 16#ffff (halfword (wplus2 block 22) 0))))
     (date nil . ,(setq da (halfword (wplus2 block 24) 0)))
     (accesstime nil .
	 ,(setq tm (sys2int (wor (wshift da 16) tm))))  % UNIX style
     (writetime  nil . ,tm)
     (size nil . ,(getmem (wplus2 block 26)))
     (name nil . ,(compress (reverse n))) )))
 
(lap '((*entry set-dta expr 1)
       (*move (reg 1) (reg 4))
       (*move (times 256 16#1A) (reg 1))
       (int 16#21)
       (*exit 0)))
 
(lap '((*entry get-dta expr 0)
       (*move (times 256 16#2F) (reg 1))
       (int 16#21)
       (*move (reg 2)(reg 1))
       (*exit 0)))
 
(lap '((*entry find-first expr 1)
       (*move (reg 1)(reg 4))
       (*move 16#10 (reg 3))              % including directory search
       (*move (times 256 16#4e) (reg 1))
       (int 16#21)
       (*exit 0) ))
 
(lap '((*entry find-next expr 0)
       (*move (times 256 16#4f) (reg 1))
       (int 16#21)
       (*exit 0) ))

(de directory(u)
  (when (weq **windows 1) (windows-error))       
  (let (d r)
    (set-dta filestatus-work)
    (setq u (if (and (stringp u) (wgreaterp (strlen (strinf u)) 0))
		(fnexpand u)
		"*.*"))
    (setq r (find-first (strbase (strinf u))))
    (while (weq r 0)
	  (push (cddr(atsoc 'name (expand-status-block))) d)
	  (setq r (find-next)) )
    (setq d (reversip d))
    d
    ))
 
%---------------------------------------------------------------------------

(de returnaddressp (x)
  (prog (s y)
	(unless (and (intp x) (>= x 2000))
	  (return nil))
	% Actually, top bits must                                          
	% be 0 or -1 due to                                                
	% 9836 assembler, linker                                           
	(when (weq (wand x 1) 1)
	  (return nil))
	% if OddP X                                                        
	(setq x (inf x))
	(when (wlessp x 8198)
	  (return nil))
	(cond ((not (wlessp x (inf nextbps))) % Assures X points to real memory
	       (return nil)))
	(setq s (inf symfnc))
	(unless (weq (halfword x -3) 16#15ff) (return nil))
	% call longword                                                     
	(setq y (inf (wgetv x -1)))
	(setq y (wdifference y s))
	(setq y (wquotient y addressingunitsperfunctioncell))
	(if (or (wlessp y 0) (wgreaterp y maxsymbols))
	  (return nil)
	  (return (mkid y)))))

% -----------------------------------------------------------------        
(fluid '(channeltable))

(de channelflush (chnl)
  % Flush any channel. 
  (fflush (wgetv channeltable chnl)))

% ----------------------------------------------------------------         
% String-oriented Unix interface functions.                                

% Copy and tag a Lisp string, given a C string pointer.                    
(de importforeignstring (c_s)
  (prog (new_s len)
	(when (weq c_s 0)
	  (return nil))
	% Not a string, pass it on. 
	(setq len (wdifference (external_strlen c_s) 1))
	(setq new_s (gtstr len))
	(for (from i 0 len 1) 
	      (do (setf (strbyt new_s i) (byte c_s i))))
	(return (mkstr new_s))))

(de external-allocatemorebps ()
  (allocatemorebps))

(de init-file-string (program-name)
  % Build init file name: append RC to the name
  (bldmsg "%w%w.rc" (user-homedir-string) program-name))

(de user-homedir-string () 
  (let ((x (getenv "HOME")))
       (if x (bldmsg "%w\" x) "")))

(de getenv (s)
  % String from environment, or NIL.                                    
  (if (stringp s) (envexpand s) nil))

(de setenv (var val) (stderror "setenv not implemented"))

(de cd (s)                             
   % Set current working directory.
  (when (idp s) (setq s (id2string s)))
  (when (stringp s)
   (weq 0 (unixcd (strbase (strinf (fnexpand s)))))))     % 0 is success.

%---------- expand a filename from the environment ---------------------

(fluid '(psp-sel testenv))

(compiletime (put 'dollar 'CharConst 36))

(de envexpand (s)
    (setq s (errorset `(fnexpand1 ,(bldmsg "$%w" s) nil) nil nil))
    (if (pairp s) (car s) nil))

(de fnexpand (na) (fnexpand1 na t))

(compiletime (flag '(spec-charp) 'internalentry))

(de spec-charp(c)
  % test whether c is not a character and not a digit
 (not
  (or (and (wleq (char 'a) c)(wleq c (char 'z)))
      (and (wleq (char '!a) c)(wleq c (char '!z))) 
      (and (wleq (char '0) c)(wleq c (char '9)))))) 
      
(de fnexpand1(na mode)
 (if (or (null psp-sel) (not(stringp na))) na
  (let ((lna (strlen (strinf na)))
	(envseg (halfword_es 44 psp-sel))
	dpos lpos envpos i j lvar lsub c clist res
       )
     (prog()
      (ifor(from i 0 lna 1)
	   (do (when (eq (strbyt (strinf na) i) (char dollar))
		     (setq dpos (iadd1 i)) )))
      (when (null dpos)(return na))
      (setq envpos 0)
   test-name
      (setq i 0)
     next-char
      (setq c (byte_es (wplus2 envpos i) envseg))
      (cond((eq c 1) (go error))
	   ((and(eq c (char =)) 
		(spec-charp (strbyt (strinf na) (wplus2 dpos i))))
		(go found))
	   ((eq (upper-case c)
		(upper-case (strbyt (strinf na) (wplus2 dpos i))))
	    (setq i (iadd1 i))
	    (when (wgreaterp (wplus2 dpos i) (iadd1 lna))
		  (go next-name))
	    (go next-char))
	   (t (go next-name)) )
   next-name
      (while (and (not (eq c 0))(not (eq c 1)))
	     (setq envpos (iadd1 envpos))
	     (setq c (byte_es envpos envseg)))
      (when (eq c 1) (go error))
      (setq envpos (iadd1 envpos))
      (go test-name)

   error
      (when (null mode)
       (return (stderror
		(bldmsg "cannot expand >%w< from environment" na))))
      (setq na (explode na))
      (when (setq c (member '!$ na))(rplaca c '!\))
      (return (fnexpand (compress na)))
   found
	   % i points to the equal sign; extract target string
      (setq lvar i)  % save length of variable
      (setq i(wplus2 envpos i))
      (while (and (not (eq (setq c (byte_es (setq i(iadd1 i)) envseg)) 0))
		  (not (eq c 1)))
	     (setq clist (cons c clist)))
      (setq lsub (length clist))
      (setq clist (reversip clist))
	   % collect result string
      (setq res (mkstring (isub1 (wplus2 (wdifference lna lvar)
				 lsub))))
      (when (wgreaterp dpos 1)
       (ifor (from i 0 (idifference dpos 2) 1)
	    (do (setf (strbyt (strinf res) i)
		      (strbyt (strinf na) i)) )))
       (setq j (isub1 dpos))
       (while clist (setf (strbyt (strinf res) j) (car clist))
		    (setq j (iadd1 j))
		    (setq clist (cdr clist)))
       (ifor (from i (wplus2 dpos lvar) lna 1)
	    (do (progn
		  (setf (strbyt (strinf res) j)
			(strbyt (strinf na) i))
		  (setq j (iadd1 j)) )))
	  (setf (strbyt (strinf res) j) 0)  % trailing zero
      (return (fnexpand res))
      ))))


%(de unix-time ()
%  (sys2int (external_time 0)))

(de &time-control (u) nil)  % hook

(setq loaddirectories* '( "" ))

%---------------------------------------------------------------------
%
%   DOS command interface

(fluid '(
     command-handle    % internal number for rpc 
	 command-slave     % full path name of slave (.exe-file)
	 ))

(compiletime (flag '(dos-handle rp-handle rp-call )
		   'internalfunction))
    
(lap '((*entry rp-handle expr 1)
       (*move (reg 1)($fluid name))
       (*move (reg ds)(reg 2))
       (*move (reg 2)(reg es))
       (*move (reg 1)(reg edx))
       (*move 16#e000 (reg 1))
       (int 16#21)
	   (*wand (reg 1) 255)
	   (*exit 0)))

(lap '((*entry rp-call expr 3)
       (*wor (reg 1) 16#e100)
       (*move (reg 2) (reg edx))   % parameter string
       (*move (reg 3)(reg 3))      % lth of string
       (*move 0 (reg 2))           % result length
       (int 16#21)
       (*exit 0) ))
 
(de rcall-init(name)
   (when (weq **windows 1) (windows-error))       
   (rp-handle (strbase(strinf name))))

(de rcall(handle par)
   (when (weq **windows 1) (windows-error))       
   (setq par
      (cond
	     ((stringp par)
	      (bldmsg "%c%w  $  $" (iadd1 (strlen (strinf par))) par))
		 ((null par)
		  (bldmsg "%c    $  $" 0))  % empty command line
		 (t (stderror " rcall parameter error"))))
   (rp-call handle
	(strbase (strinf par))
	(wplus2 5 (strlen (strinf par)))))
 
(de dos-handle()
   (when (weq **windows 1) (windows-error))
   (when (not (stringp command-slave))
	 (stderror "command slave filename missing"))
   (setq command-handle
	(or command-handle
	    (rcall-init (fnexpand command-slave))))
   (cond ((eq command-handle 2)
	  (stderror "command slave file not found"))
		 ((eq command-handle 8)
		  (stderror "not enough memory for command slave exec")))
   command-handle)

(dm system(u) `(system1 ,(if (cdr u) (cadr u) nil )))

(de system1(command)
  (&time-control nil)
  (if (weq **windows 1) 
      (when (stringp command)
	    (psll-call (strbase(strinf (fnexpand command))) 0 0 17))
      (rcall (dos-handle)
	  (if (not (stringp command))
		      ""
			  (fnexpand (bldmsg "/c %w" command) )))
       t)
   (&time-control t)
)

(de windows-error()
      (stderror "function not available under windows"))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Access to some BIOS functions

(lap '((*entry bios-int11h expr 0)
	      % read configuration word
       (int 16#11)
	   (*exit 0)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Handling priority messages from Windows
%

(compiletime (load fast-vectors))

(fluid '(win-messages))

(setq win-messages (make-vector 31 nil))
 
(iputv win-messages 1 '(channellinelength 1 (psll-call 0 0 0 20))) 

(de prio-message1(n)
 (ifor (from i 0 31 1)
    (do
     (progn
       (when (wneq 0 (wand n 1)) (eval (igetv win-messages i)))
       (setq n (wshift n -1))
))  ))

% enable windows interrupts
(de ! yield()(psll-call 0 0 0 0))

% prompt an item (returns a string)
(de askUser(q)
   (let(a c n)
     (cond ((weq **windows 1)
	    (setq c (gtwrds 10))   % buffer for answer
	    (setq n
	       (psll-call (strbase(strinf q))
			  (strbase(strinf c))
			  40
			  21))
	       % convert string to list of chars
	    (ifor (from i 0 (isub1 n) 1)
		  (do (push (int2id (strbyt (strinf c) i)) a)))
	   )
	   (t (let ((out* 1)(in* 0))
		   (prin2t q)
		   (while (wneq (setq c(readch)) (char eol))
			  (push c a))
	   )  )
      )
      (if (null a)
	  nil
	  (compress (cons '!" (reverse (cons '!" a))))
    )))
       
	 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  General interface for message passing DDE for windows
%
%  H. Melenk, ZIB Berlin, January 1992
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(*message-buffer* *token-buffer* *callback-functions*))
 
(setq *message-buffer* (mkstring 128 0)
      *token-buffer* (mkstring 128 0))
 
(de send-server(handle command text)
   (test-server handle 'send-server)
   (when (not (stringp command))
	 (nonstringerror command 'send-server))
   (when (not (stringp text))
	 (nonstringerror text 'send-server))
   (psll-call handle (strbase (strinf command))
		      (strbase (strinf text))
		      CONNECT_SEND))
 
(de fetch-server(handle command)
  (prog (l n r)
   (test-server handle 'fetch-server)
   (when (not (stringp command))
	 (nonstringerror command 'fetch-server))
   (setf (strbyt (strinf *message-buffer*) 0) 0)
   (setq n
    (psll-call handle (strbase (strinf command))
		      (strbase (strinf *message-buffer*))
		       CONNECT_FETCH))
   (when (or (wleq n 0)
	     (weq 0 (strbyt (strinf *message-buffer*) 0) ))
	 (return nil))
   (setq r (copy-message *message-buffer*))
   (return r)))
 
 
(de open-server(server topic callback)
 (prog(u)
   (when (not (stringp server))
	 (nonstringerror server 'open-server))
   (when (not (stringp topic))
	 (nonstringerror topic 'open-server))
   (setq u
    (psll-call (strbase (strinf server)) (strbase (strinf topic))
	       0 CONNECT_OPEN))
   (when (wleq u 0) (return nil))
   (when callback (push (cons u callback) *callback-functions*))
   (return u)))
 
 
(de close-server(handle)
     (delasc handle *callback-functions*)
     (psll-call handle 0 0 CONNECT_CLOSE))
 
(de message-handler()
    % this routine is called when an asynchronous message has arrived;
    % action: the handle, message type and message text picked up;
    %         if there is a callback function for the handle, this
    %         is executed with the message tag and text as arguments.
  (prog(hand cb tag)
    (setq hand
       (PSLL-call (strbase (strinf *token-buffer*))
		  (strbase (strinf *message-buffer*))
		  0
		  CONNECT_ASK))
    (when (or (wleq hand 0) (eq tag 'ACK)) (return nil))
    (setq tag (intern (copy-message *token-buffer*)))
    (setq cb (assoc hand *callback-functions*))
    (when cb (apply (cdr cb)
		   (list tag
			 (copy-message *message-buffer*) )))
    (return nil) ))
 
(de test-server(handle fcn)
   (when (not (and (fixp handle)(wgreaterp handle 0)))
	 (typeerror handle fcn 'handle)))
 
(de copy-message(msg)
 % copy characters from global buffer to local string.
 (prog(l r)
   (setq l 0)
   (while (not (izerop (strbyt (strinf msg) l)))
	  (setq l (iadd1 l)))
   (setq l (isub1 l))
   (setq r (mkstring l))
   (ifor (from i 0 l 1)
	 (do (setf (strbyt (strinf r) i)
		   (strbyt (strinf msg) i))))
   (return r) ))

(de sleep(n)
  (when (eq **windows 1)  
    (psll-call n 0 0 SLEEP)))

(de psll-reset(n)
  (when (eq **windows 1)  
    (psll-call n 0 0 PSLL_RESET)))



(off fast-integers)

%% End of File.
